#!/usr/bin/env python
# --!-- coding: utf8 --!--
import re

from PyQt5.QtGui import QTextCharFormat, QFont
from PyQt5.QtWidgets import qApp, QVBoxLayout, QCheckBox, QWidget, QHBoxLayout, QLabel, QSpinBox, QComboBox

from manuskript.exporter.manuskript.markdown import markdown, markdownSettings
from manuskript.ui.collapsibleGroupBox2 import collapsibleGroupBox2


class abstractPlainText(markdown):
    name = "SUBCLASSME"
    description = "SUBCLASSME"
    exportVarName = "SUBCLASSME"
    toFormat = "SUBCLASSME"
    icon = "SUBCLASSME"
    exportFilter = "SUBCLASSME"
    exportDefaultSuffix = ".SUBCLASSME"

    def __init__(self, exporter):
        self.exporter = exporter

    def settingsWidget(self):
        # Get pandoc major version to determine valid command line options
        p = re.compile(r'pandoc (\d+)\..*')
        m = p.match(self.exporter.version())
        if m:
            majorVersion = m.group(1)
        else:
            majorVersion = ""
        w = pandocSettings(self, majorVersion, toFormat=self.toFormat)
        w.loadSettings()
        return w

    def src(self, settingsWidget):
        return markdown.output(self, settingsWidget)

    def output(self, settingsWidget, outputfile=None):
        args = settingsWidget.runnableSettings()
        src = self.src(settingsWidget)
        return self.exporter.convert(src, args, outputfile)

    def preview(self, settingsWidget, previewWidget):
        settings = settingsWidget.getSettings()

        # Save settings
        settingsWidget.writeSettings()

        # Prepares text edit
        self.preparesTextEditViewMarkdown(previewWidget, settingsWidget.settings)
        self.preparesTextEditView(previewWidget, settings["Preview"]["PreviewFont"])

        r = self.output(settingsWidget)

        previewWidget.setPlainText(r)


class pandocSetting:
    def __init__(self, arg, type, format, label, widget=None, default=None, min=None, max=None, vals=None, suffix=""):
        self.arg = arg  # start with EXT for extensions
        self.type = type
        self.label = label
        self.formats = format
        if "html" in self.formats:
            self.formats += " slidy dzslides revealjs slideous s5 html5"
        self.widget = widget
        self.default = default
        self.min = min
        self.max = max
        self.vals = vals.split("|") if vals else []
        self.suffix = suffix

    def isValid(self, format):
        """Return whether the specific setting is active with the given format."""

        # Empty formats means all
        if self.formats is "":
            return True

        # "html" in "html markdown latex"
        elif format in self.formats:
            return True

        # "markdown_strict" in "html markdown latex"
        elif [f for f in self.formats.split(" ") if format in f]:
            return True

        return False


class pandocSettings(markdownSettings):

    settingsList = {
        # General
        "standalone": pandocSetting("--standalone", "checkbox", "",
                                    qApp.translate("Export", "Standalone document (not just a fragment)"),
                                    default=True),
        "TOC":          pandocSetting("--toc", "checkbox", "",
                                      qApp.translate("Export", "Include a table of contents.")),

        "TOC-depth":    pandocSetting("--toc-depth=", "number", "",
                                      qApp.translate("Export", "Number of sections level to include in TOC: "),
                                      default=3, min=1, max=6),
        # pandoc v1 only
        "smart":        pandocSetting("--smart", "checkbox", "",
                                      qApp.translate("Export", "Typographically correct output")),
        # pandoc v1 only
        "normalize":    pandocSetting("--normalize", "checkbox", "",
                                      qApp.translate("Export", "Normalize the document (cleaner)")),
        "base-header":  pandocSetting("--base-header-level=", "number", "",
                                      qApp.translate("Export", "Specify the base level for headers: "),
                                      default=1, min=1),
        "disable-YAML": pandocSetting("EXT-yaml_metadata_block", "checkbox", "",
                                      qApp.translate("Export", "Disable YAML metadata block.\nUse that if you get YAML related error.")),

        # Specific
        "ref-link":     pandocSetting("--reference-links", "checkbox", "markdown rst",
                                    qApp.translate("Export", "Use reference-style links instead of inline links")),
        "atx":          pandocSetting("--atx-headers", "checkbox", "markdown asciidoc",
                                    qApp.translate("Export", "Use ATX-style headers")),
        "self-contained": pandocSetting("--self-contained", "checkbox", "html",
                                        qApp.translate("Export", "Self-contained HTML files, with no dependencies")),
        "q-tags":       pandocSetting("--html-q-tags", "checkbox", "html",
                                        qApp.translate("Export", "Use <q> tags for quotes in HTML")),
        # pandoc v1 only
        "latex-engine": pandocSetting("--latex-engine=", "combo", "pdf",
                                      qApp.translate("Export", "LaTeX engine used to produce the PDF."),
                                      vals="pdflatex|lualatex|xelatex"),
        # pandoc v2
        "pdf-engine":   pandocSetting("--pdf-engine=", "combo", "pdf",
                                      qApp.translate("Export", "LaTeX engine used to produce the PDF."),
                                      vals="pdflatex|lualatex|xelatex"),
        "epub3":        pandocSetting("EXTepub3", "checkbox", "epub",
                                        qApp.translate("Export", "Convert to ePUB3")),
    }

    pdfSettings = {

        # PDF
        "latex-ps":     pandocSetting("--variable=papersize:", "combo", "pdf latex",  # FIXME: does not work with default template
                                      qApp.translate("Export", "Paper size:"),
                                      vals="letter|A4|A5"),
        "latex-fs":     pandocSetting("--variable=fontsize:", "number", "pdf latex",  # FIXME: does not work with default template
                                      qApp.translate("Export", "Font size:"),
                                      min=8, max=88, default=12, suffix="pt"),
        "latex-class":  pandocSetting("--variable=documentclass:", "combo", "pdf latex",
                                     qApp.translate("Export", "Class:"),
                                     vals="article|report|book|memoir"),
        "latex-ls":     pandocSetting("--variable=linestretch:", "combo", "pdf latex",
                                     qApp.translate("Export", "Line spacing:"),
                                     vals="1|1.25|1.5|2"),

        # FIXME: complete with http://pandoc.org/README.html#variables-for-latex
    }


    def __init__(self, _format, majorVersion="", toFormat=None, parent=None):
        markdownSettings.__init__(self, _format, parent)

        self.format = toFormat
        self.majorVersion = majorVersion

        w = QWidget(self)
        w.setLayout(QVBoxLayout())
        self.grpPandocGeneral = self.collapsibleGroupBox(self.tr("General"), w)

        if majorVersion == "1":
            # pandoc v1 only
            self.addSettingsWidget("smart", self.grpPandocGeneral)
            self.addSettingsWidget("normalize", self.grpPandocGeneral)
        else:
            # pandoc v2
            self.settingsList.pop("smart", None)
            self.settingsList.pop("normalize", None)
        self.addSettingsWidget("base-header", self.grpPandocGeneral)
        self.addSettingsWidget("standalone", self.grpPandocGeneral)
        self.addSettingsWidget("disable-YAML", self.grpPandocGeneral)

        self.grpPandocTOC = self.collapsibleGroupBox(self.tr("Table of Content"), w)

        self.addSettingsWidget("TOC", self.grpPandocTOC)
        self.addSettingsWidget("TOC-depth", self.grpPandocTOC)

        self.grpPandocSpecific = self.collapsibleGroupBox(self.tr("Custom settings for {}").format(self.format), w)

        self.addSettingsWidget("ref-link", self.grpPandocSpecific)
        self.addSettingsWidget("atx", self.grpPandocSpecific)
        self.addSettingsWidget("self-contained", self.grpPandocSpecific)
        self.addSettingsWidget("q-tags", self.grpPandocSpecific)
        if majorVersion == "1":
            # pandoc v1 only
            self.addSettingsWidget("latex-engine", self.grpPandocSpecific)
            self.settingsList.pop("pdf-engine", None)
        else:
            # pandoc v2
            self.settingsList.pop("latex-engine", None)
            self.addSettingsWidget("pdf-engine", self.grpPandocSpecific)
        self.addSettingsWidget("epub3", self.grpPandocSpecific)

        # PDF settings
        self.settingsList.update(self.pdfSettings)
        for i in self.pdfSettings:
            self.addSettingsWidget(i, self.grpPandocSpecific)

        self.toolBox.insertItem(self.toolBox.count() - 1, w, "Pandoc")
        self.toolBox.layout().setSpacing(0)  # Not sure why this is needed, but hey...

        self.getSettings()

    def collapsibleGroupBox(self, title, parent):
        g = collapsibleGroupBox2(title=title)
        parent.layout().addWidget(g)
        g.setLayout(QVBoxLayout())
        return g

    def addSettingsWidget(self, settingsName, parent):

        if not settingsName in self.settingsList:
            return

        s = self.settingsList[settingsName]

        if not s.isValid(self.format):
            # That setting is not available for that export format
            return

        if "checkbox" in s.type:
            s.widget = QCheckBox(s.label)
            if s.default:
                s.widget.setChecked(True)
            parent.layout().addWidget(s.widget)

        elif "number" in s.type:
            l = QHBoxLayout()
            label = QLabel(s.label, parent)
            label.setWordWrap(True)
            l.addWidget(label, 8)
            s.widget = QSpinBox()
            s.widget.setValue(s.default if s.default else 0)
            if s.min:
                s.widget.setMinimum(s.min)
            if s.max:
                s.widget.setMaximum(s.max)
            if s.suffix:
                s.widget.setSuffix(s.suffix)
            l.addWidget(s.widget, 2)
            parent.layout().addLayout(l)

        elif "combo" in s.type:
            l = QHBoxLayout()
            label = QLabel(s.label, parent)
            label.setWordWrap(True)
            l.addWidget(label, 6)
            s.widget = QComboBox()
            s.widget.addItems(s.vals)
            l.addWidget(s.widget, 2)
            parent.layout().addLayout(l)

    def updateFromSettings(self):
        markdownSettings.updateFromSettings(self)

        # s = self.settings["Preview"]
        # val = s.get("MarkdownHighlighter", False)
        # self.chkMarkdownHighlighter.setChecked(val)

        if not "Pandoc" in self.settings:
            return

        for name in self.settingsList:
            s = self.settingsList[name]

            if s.isValid(self.format):
                if s.type == "checkbox" and name in self.settings["Pandoc"]:
                    s.widget.setChecked(self.settings["Pandoc"][name])
                elif s.type == "number" and name in self.settings["Pandoc"]:
                    s.widget.setValue(int(self.settings["Pandoc"][name]))
                elif s.type == "combo" and name in self.settings["Pandoc"]:
                    s.widget.setCurrentText(self.settings["Pandoc"][name])


    def getSettings(self):
        self.settings = markdownSettings.getSettings(self)

        P = self.settings.get("Pandoc", {})

        for name in self.settingsList:
            s = self.settingsList[name]

            if s.isValid(self.format):
                if s.type == "checkbox":
                    P[name] = s.widget.isChecked()
                elif s.type == "number":
                    P[name] = str(s.widget.value())
                elif s.type == "combo":
                    P[name] = s.widget.currentText()

        self.settings["Pandoc"] = P

        return self.settings

    def runnableSettings(self):

        # First we get extensions (where arg starts with EXT)
        extensions = ""
        toFormat = self.format
        for name in self.settingsList:
            s = self.settingsList[name]
            if s.arg[:3] == "EXT" and s.isValid(self.format):
                if name == "disable-YAML" and s.widget.isChecked():
                    extensions += "-yaml_metadata_block"
                if name == "epub3" and s.widget.isChecked():
                    toFormat = "epub3"

        r = ["--from=markdown" + extensions,
             "--to={}".format(toFormat)]

        # Add every command
        for name in self.settingsList:
            s = self.settingsList[name]

            if s.arg[:3] == "EXT":
                continue

            if s.isValid(self.format):
                rr = ""
                if s.type == "checkbox":
                    if s.widget.isChecked():
                        rr = s.arg
                elif s.type == "number":
                    rr = "{}{}".format(
                        s.arg,
                        str(s.widget.value())
                    )
                elif s.type == "combo":
                    rr = "{}{}".format(
                        s.arg,
                        s.widget.currentText()
                    )

                if rr:
                    r.append(rr+s.suffix)
        return r




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'manuskript/ui/importers/generalSettings_ui.ui'
#
# Created by: PyQt5 UI code generator 5.9
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_generalSettings(object):
    def setupUi(self, generalSettings):
        generalSettings.setObjectName("generalSettings")
        generalSettings.resize(267, 401)
        generalSettings.setWindowTitle("Form")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(generalSettings)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setSpacing(10)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.toolBox = QtWidgets.QToolBox(generalSettings)
        self.toolBox.setObjectName("toolBox")
        self.general = QtWidgets.QWidget()
        self.general.setGeometry(QtCore.QRect(0, 0, 267, 375))
        self.general.setObjectName("general")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.general)
        self.verticalLayout_5.setContentsMargins(6, 6, 6, 6)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.formLayout_4 = QtWidgets.QFormLayout()
        self.formLayout_4.setRowWrapPolicy(QtWidgets.QFormLayout.WrapLongRows)
        self.formLayout_4.setObjectName("formLayout_4")
        self.chkGeneralSplitScenes = QtWidgets.QCheckBox(self.general)
        self.chkGeneralSplitScenes.setObjectName("chkGeneralSplitScenes")
        self.formLayout_4.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.chkGeneralSplitScenes)
        self.txtGeneralSplitScenes = QtWidgets.QLineEdit(self.general)
        self.txtGeneralSplitScenes.setText("")
        self.txtGeneralSplitScenes.setObjectName("txtGeneralSplitScenes")
        self.formLayout_4.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.txtGeneralSplitScenes)
        self.chkGeneralTrimTitles = QtWidgets.QCheckBox(self.general)
        self.chkGeneralTrimTitles.setObjectName("chkGeneralTrimTitles")
        self.formLayout_4.setWidget(4, QtWidgets.QFormLayout.SpanningRole, self.chkGeneralTrimTitles)
        self.treeGeneralParent = QtWidgets.QTreeView(self.general)
        self.treeGeneralParent.setHeaderHidden(True)
        self.treeGeneralParent.setObjectName("treeGeneralParent")
        self.formLayout_4.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.treeGeneralParent)
        self.chkGeneralParent = QtWidgets.QCheckBox(self.general)
        self.chkGeneralParent.setObjectName("chkGeneralParent")
        self.formLayout_4.setWidget(0, QtWidgets.QFormLayout.SpanningRole, self.chkGeneralParent)
        self.chkGeneralTopLevel = QtWidgets.QCheckBox(self.general)
        self.chkGeneralTopLevel.setObjectName("chkGeneralTopLevel")
        self.formLayout_4.setWidget(2, QtWidgets.QFormLayout.SpanningRole, self.chkGeneralTopLevel)
        self.verticalLayout_5.addLayout(self.formLayout_4)
        self.toolBox.addItem(self.general, "")
        self.verticalLayout_2.addWidget(self.toolBox)

        self.retranslateUi(generalSettings)
        self.toolBox.setCurrentIndex(0)
        self.toolBox.layout().setSpacing(0)
        QtCore.QMetaObject.connectSlotsByName(generalSettings)

    def retranslateUi(self, generalSettings):
        _translate = QtCore.QCoreApplication.translate
        self.chkGeneralSplitScenes.setText(_translate("generalSettings", "Split scenes at:"))
        self.txtGeneralSplitScenes.setPlaceholderText(_translate("generalSettings", "\\n---\\n"))
        self.chkGeneralTrimTitles.setText(_translate("generalSettings", "Trim long titles (> 32 chars)"))
        self.chkGeneralParent.setText(_translate("generalSettings", "Import under:"))
        self.chkGeneralTopLevel.setText(_translate("generalSettings", "Import in a top-level folder"))
        self.toolBox.setItemText(self.toolBox.indexOf(self.general), _translate("generalSettings", "General"))




#!/usr/bin/env python
# --!-- coding: utf8 --!--

import datetime
import difflib
import re

from PyQt5.QtCore import Qt, QTimer, QRect
from PyQt5.QtGui import QPalette, QFontMetrics
from PyQt5.QtWidgets import QWidget, QMenu, QActionGroup, QAction, QListWidgetItem, QStyledItemDelegate, QStyle

from manuskript.enums import Outline
from manuskript.ui.revisions_ui import Ui_revisions
from manuskript.models import references as Ref


class revisions(QWidget, Ui_revisions):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.setupUi(self)
        self.splitter.setStretchFactor(0, 5)
        self.splitter.setStretchFactor(1, 70)

        self.listDelegate = listCompleterDelegate(self)
        self.list.setItemDelegate(self.listDelegate)
        self.list.itemClicked.connect(self.showDiff)
        self.list.setContextMenuPolicy(Qt.CustomContextMenu)
        self.list.customContextMenuRequested.connect(self.popupMenu)
        self.btnDelete.setEnabled(False)
        self.btnDelete.clicked.connect(self.delete)
        self.btnRestore.clicked.connect(self.restore)
        self.btnRestore.setEnabled(False)

        # self.list.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.updateTimer = QTimer()
        self.updateTimer.setSingleShot(True)
        self.updateTimer.setInterval(500)
        self.updateTimer.timeout.connect(self.update)
        self.updateTimer.stop()

        self.menu = QMenu(self)
        self.actGroup = QActionGroup(self)

        self.actShowDiff = QAction(self.tr("Show modifications"), self.menu)
        self.actShowDiff.setCheckable(True)
        self.actShowDiff.setChecked(True)
        self.actShowDiff.triggered.connect(self.showDiff)
        self.menu.addAction(self.actShowDiff)
        self.actGroup.addAction(self.actShowDiff)

        self.actShowVersion = QAction(self.tr("Show ancient version"), self.menu)
        self.actShowVersion.setCheckable(True)
        self.actShowVersion.setChecked(False)
        self.actShowVersion.triggered.connect(self.showDiff)
        self.menu.addAction(self.actShowVersion)
        self.actGroup.addAction(self.actShowVersion)

        self.menu.addSeparator()
        self.actShowSpaces = QAction(self.tr("Show spaces"), self.menu)
        self.actShowSpaces.setCheckable(True)
        self.actShowSpaces.setChecked(False)
        self.actShowSpaces.triggered.connect(self.showDiff)
        self.menu.addAction(self.actShowSpaces)

        self.actDiffOnly = QAction(self.tr("Show modifications only"), self.menu)
        self.actDiffOnly.setCheckable(True)
        self.actDiffOnly.setChecked(True)
        self.actDiffOnly.triggered.connect(self.showDiff)
        self.menu.addAction(self.actDiffOnly)
        self.btnOptions.setMenu(self.menu)

        self._model = None
        self._index = None

    def setModel(self, model):
        self._model = model
        self._model.dataChanged.connect(self.updateMaybe)

    def setCurrentModelIndex(self, index):
        self._index = index
        self.view.setText("")
        self.update()

    def updateMaybe(self, topLeft, bottomRight):
        if self._index and \
                                topLeft.column() <= Outline.revisions <= bottomRight.column() and \
                                topLeft.row() <= self._index.row() <= bottomRight.row():
            # self.update()
            self.updateTimer.start()

    def update(self):
        self.list.clear()
        item = self._index.internalPointer()
        rev = item.revisions()
        # Sort revisions
        rev = sorted(rev, key=lambda x: x[0], reverse=True)
        for r in rev:
            timestamp = datetime.datetime.fromtimestamp(r[0]).strftime('%Y-%m-%d %H:%M:%S')
            readable = self.readableDelta(r[0])
            i = QListWidgetItem(readable)
            i.setData(Qt.UserRole, r[0])
            i.setData(Qt.UserRole + 1, timestamp)
            self.list.addItem(i)

    def readableDelta(self, timestamp):
        now = datetime.datetime.now()
        delta = now - datetime.datetime.fromtimestamp(timestamp)
        if delta.days > 365:
            return self.tr("{} years ago").format(str(int(delta.days / 365)))
        elif delta.days > 30:
            return self.tr("{} months ago").format(str(int(delta.days / 30.5)))
        elif delta.days > 0:
            return self.tr("{} days ago").format(str(delta.days))
        if delta.days == 1:
            return self.tr("1 day ago")
        elif delta.seconds > 60 * 60:
            return self.tr("{} hours ago").format(str(int(delta.seconds / 60 / 60)))
        elif delta.seconds > 60:
            return self.tr("{} minutes ago").format(str(int(delta.seconds / 60)))
        else:
            return self.tr("{} seconds ago").format(str(delta.seconds))

    def showDiff(self):
        # UI stuff
        self.actShowSpaces.setEnabled(self.actShowDiff.isChecked())
        self.actDiffOnly.setEnabled(self.actShowDiff.isChecked())

        # FIXME: Errors in line number
        i = self.list.currentItem()

        if not i:
            self.btnDelete.setEnabled(False)
            self.btnRestore.setEnabled(False)
            return

        self.btnDelete.setEnabled(True)
        self.btnRestore.setEnabled(True)

        ts = i.data(Qt.UserRole)
        item = self._index.internalPointer()

        textNow = item.text()
        textBefore = [r[1] for r in item.revisions() if r[0] == ts][0]

        if self.actShowVersion.isChecked():
            self.view.setText(textBefore)
            return

        textNow = textNow.splitlines()
        textBefore = textBefore.splitlines()

        d = difflib.Differ()
        diff = list(d.compare(textBefore, textNow))

        if self.actShowSpaces.isChecked():
            _format = lambda x: x.replace(" ", "␣ ")
        else:
            _format = lambda x: x

        extra = "<br>"
        diff = [d for d in diff if d and not d[:2] == "? "]
        mydiff = ""
        skip = False
        for n, l in enumerate(diff):
            l = diff[n]
            op = l[:2]
            txt = l[2:]
            op2 = diff[n + 1][:2] if n + 1 < len(diff) else None
            txt2 = diff[n + 1][2:] if n + 1 < len(diff) else None

            if skip:
                skip = False
                continue

            # Same line
            if op == "  " and not self.actDiffOnly.isChecked():
                mydiff += "{}{}".format(txt, extra)

            elif op == "- " and op2 == "+ ":
                if self.actDiffOnly.isChecked():
                    mydiff += "<br><span style='color: blue;'>{}</span><br>".format(
                            self.tr("Line {}:").format(str(n)))
                s = difflib.SequenceMatcher(None, txt, txt2, autojunk=True)
                newline = ""
                for tag, i1, i2, j1, j2 in s.get_opcodes():
                    if tag == "equal":
                        newline += txt[i1:i2]
                    elif tag == "delete":
                        newline += "<span style='color:red; background:yellow;'>{}</span>".format(_format(txt[i1:i2]))
                    elif tag == "insert":
                        newline += "<span style='color:green; background:yellow;'>{}</span>".format(
                            _format(txt2[j1:j2]))
                    elif tag == "replace":
                        newline += "<span style='color:red; background:yellow;'>{}</span>".format(_format(txt[i1:i2]))
                        newline += "<span style='color:green; background:yellow;'>{}</span>".format(
                            _format(txt2[j1:j2]))

                # Few ugly tweaks for html diffs
                newline = re.sub(r"(<span style='color.*?><span.*?>)</span>(.*)<span style='color:.*?>(</span></span>)",
                                 "\\1\\2\\3", newline)
                newline = re.sub(
                    r"<p align=\"<span style='color:red; background:yellow;'>cen</span><span style='color:green; background:yellow;'>righ</span>t<span style='color:red; background:yellow;'>er</span>\" style=\" -qt-block-indent:0; -qt-user-state:0; \">(.*?)</p>",
                    "<p align=\"right\"><span style='color:green; background:yellow;'>\\1</span></p>", newline)
                newline = re.sub(
                    r"<p align=\"<span style='color:green; background:yellow;'>cente</span>r<span style='color:red; background:yellow;'>ight</span>\" style=\" -qt-block-indent:0; -qt-user-state:0; \">(.*)</p>",
                    "<p align=\"center\"><span style='color:green; background:yellow;'>\\1</span></p>", newline)
                newline = re.sub(r"<p(<span.*?>)(.*?)(</span>)(.*?)>(.*?)</p>",
                                 "<p\\2\\4>\\1\\5\\3</p>", newline)

                mydiff += newline + extra
                skip = True
            elif op == "- ":
                if self.actDiffOnly.isChecked():
                    mydiff += "<br>{}:<br>".format(str(n))
                mydiff += "<span style='color:red;'>{}</span>{}".format(txt, extra)
            elif op == "+ ":
                if self.actDiffOnly.isChecked():
                    mydiff += "<br>{}:<br>".format(str(n))
                mydiff += "<span style='color:green;'>{}</span>{}".format(txt, extra)

        self.view.setText(mydiff)

    def restore(self):
        i = self.list.currentItem()
        if not i:
            return
        ts = i.data(Qt.UserRole)
        item = self._index.internalPointer()
        textBefore = [r[1] for r in item.revisions() if r[0] == ts][0]
        index = self._index.sibling(self._index.row(), Outline.text)
        self._index.model().setData(index, textBefore)
        # item.setData(Outline.text, textBefore)

    def delete(self):
        i = self.list.currentItem()
        if not i:
            return
        ts = i.data(Qt.UserRole)
        self._index.internalPointer().deleteRevision(ts)

    def clearAll(self):
        self._index.internalPointer().clearAllRevisions()

    def saveState(self):
        return [
            self.actShowDiff.isChecked(),
            self.actShowVersion.isChecked(),
            self.actShowSpaces.isChecked(),
            self.actDiffOnly.isChecked(),
        ]

    def popupMenu(self, pos):
        i = self.list.itemAt(pos)
        m = QMenu(self)
        if i:
            m.addAction(self.tr("Restore")).triggered.connect(self.restore)
            m.addAction(self.tr("Delete")).triggered.connect(self.delete)
            m.addSeparator()
        if self.list.count():
            m.addAction(self.tr("Clear all")).triggered.connect(self.clearAll)

        m.popup(self.list.mapToGlobal(pos))

    def restoreState(self, state):
        self.actShowDiff.setChecked(state[0])
        self.actShowVersion.setChecked(state[1])
        self.actShowSpaces.setChecked(state[2])
        self.actDiffOnly.setChecked(state[3])
        self.actShowSpaces.setEnabled(self.actShowDiff.isChecked())
        self.actDiffOnly.setEnabled(self.actShowDiff.isChecked())


class listCompleterDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        extra = index.data(Qt.UserRole + 1)
        if not extra:
            return QStyledItemDelegate.paint(self, painter, option, index)

        else:
            if option.state & QStyle.State_Selected:
                painter.fillRect(option.rect, option.palette.color(QPalette.Inactive, QPalette.Highlight))

            title = index.data()
            extra = " - {}".format(extra)
            painter.drawText(option.rect, Qt.AlignLeft, title)

            fm = QFontMetrics(option.font)
            w = fm.width(title)
            r = QRect(option.rect)
            r.setLeft(r.left() + w)
            painter.save()
            painter.setPen(Qt.gray)
            painter.drawText(r, Qt.AlignLeft, extra)
            painter.restore()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'manuskript/ui/editors/completer_ui.ui'
#
# Created by: PyQt5 UI code generator 5.4.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtWidgets

class Ui_completer(object):
    def setupUi(self, completer):
        completer.setObjectName("completer")
        completer.resize(163, 143)
        self.verticalLayout = QtWidgets.QVBoxLayout(completer)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.text = QtWidgets.QLineEdit(completer)
        self.text.setObjectName("text")
        self.verticalLayout.addWidget(self.text)
        self.list = QtWidgets.QListWidget(completer)
        self.list.setObjectName("list")
        self.verticalLayout.addWidget(self.list)

        self.retranslateUi(completer)
        QtCore.QMetaObject.connectSlotsByName(completer)

    def retranslateUi(self, completer):
        _translate = QtCore.QCoreApplication.translate
        completer.setWindowTitle(_translate("completer", "Form"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'manuskript/ui/editors/textFormat_ui.ui'
#
# Created: Fri Apr  8 18:15:49 2016
#      by: PyQt5 UI code generator 5.2.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_textFormat(object):
    def setupUi(self, textFormat):
        textFormat.setObjectName("textFormat")
        textFormat.resize(507, 34)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(textFormat)
        self.horizontalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.format = QtWidgets.QWidget(textFormat)
        self.format.setObjectName("format")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.format)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.btnBold = QtWidgets.QToolButton(self.format)
        self.btnBold.setText("")
        self.btnBold.setObjectName("btnBold")
        self.horizontalLayout.addWidget(self.btnBold)
        self.btnItalic = QtWidgets.QToolButton(self.format)
        self.btnItalic.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnItalic.setIcon(icon)
        self.btnItalic.setObjectName("btnItalic")
        self.horizontalLayout.addWidget(self.btnItalic)
        self.btnUnderlined = QtWidgets.QToolButton(self.format)
        self.btnUnderlined.setText("")
        self.btnUnderlined.setIcon(icon)
        self.btnUnderlined.setObjectName("btnUnderlined")
        self.horizontalLayout.addWidget(self.btnUnderlined)
        self.btnClear = QtWidgets.QToolButton(self.format)
        self.btnClear.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap("../"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnClear.setIcon(icon1)
        self.btnClear.setObjectName("btnClear")
        self.horizontalLayout.addWidget(self.btnClear)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.horizontalLayout_3.addWidget(self.format)
        self.align = QtWidgets.QWidget(textFormat)
        self.align.setObjectName("align")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.align)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem2)
        self.btnLeft = QtWidgets.QToolButton(self.align)
        self.btnLeft.setText("")
        self.btnLeft.setIcon(icon)
        self.btnLeft.setObjectName("btnLeft")
        self.horizontalLayout_2.addWidget(self.btnLeft)
        self.btnCenter = QtWidgets.QToolButton(self.align)
        self.btnCenter.setText("")
        self.btnCenter.setIcon(icon)
        self.btnCenter.setObjectName("btnCenter")
        self.horizontalLayout_2.addWidget(self.btnCenter)
        self.btnRight = QtWidgets.QToolButton(self.align)
        self.btnRight.setText("")
        self.btnRight.setIcon(icon)
        self.btnRight.setObjectName("btnRight")
        self.horizontalLayout_2.addWidget(self.btnRight)
        self.btnJustify = QtWidgets.QToolButton(self.align)
        self.btnJustify.setText("")
        self.btnJustify.setIcon(icon)
        self.btnJustify.setObjectName("btnJustify")
        self.horizontalLayout_2.addWidget(self.btnJustify)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem3)
        self.horizontalLayout_3.addWidget(self.align)

        self.retranslateUi(textFormat)
        QtCore.QMetaObject.connectSlotsByName(textFormat)

    def retranslateUi(self, textFormat):
        _translate = QtCore.QCoreApplication.translate
        textFormat.setWindowTitle(_translate("textFormat", "Form"))




#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import QVariant, Qt
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtWidgets import QWidget, QTextEdit, QTableWidgetItem, QHeaderView

from manuskript import settings
from manuskript.enums import Outline
from manuskript.ui.tools.frequency_ui import Ui_FrequencyAnalyzer
import re
from collections import Counter

class frequencyAnalyzer(QWidget, Ui_FrequencyAnalyzer):
    def __init__(self, mainWindow):
        QWidget.__init__(self)
        self.setupUi(self)
        self.mw = mainWindow

        self.splitter.setSizes([10, 100])
        self.splitter.setStretchFactor(1, 10)

        self.progressBarWord.hide()
        self.tblWord.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.tblPhrase.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)


        self.btnAnalyzeWord.clicked.connect(self.analyzeWord)
        self.btnAnalyzePhrase.clicked.connect(self.analyzePhrase)

        # Settings
        self.spnWordMin.setValue(settings.frequencyAnalyzer["wordMin"])
        self.txtWordExclude.setPlainText(settings.frequencyAnalyzer["wordExclude"])
        self.spnPhraseMin.setValue(settings.frequencyAnalyzer["phraseMin"])
        self.spnPhraseMax.setValue(settings.frequencyAnalyzer["phraseMax"])
        self.spnWordMin.valueChanged.connect(self.updateSettings)
        self.txtWordExclude.textChanged.connect(self.updateSettings)
        self.spnPhraseMin.valueChanged.connect(self.updateSettings)
        self.spnPhraseMax.valueChanged.connect(self.updateSettings)

    def analyzePhrase(self):
        root = self.mw.mdlOutline.rootItem
        nMin = self.spnPhraseMin.value()
        nMax = self.spnPhraseMax.value()

        def listPhrases(item, nMin, nMax):
            txt = item.text()

            # Split into words
            lst = re.findall(r"[\w']+", txt)            # Ignores punctuation
            # lst = re.findall(r"[\w']+|[.,!?;]", txt)  # Includes punctuation
            phrases = []

            # Make tuples of n-length
            for n in range(nMin, nMax + 1):
                for l in range(len(lst) - n + 1):
                    phrases.append(tuple(lst[l:l+n]))

            for c in item.children():
                phrases += listPhrases(c, nMin, nMax)

            return phrases

        lst = listPhrases(root, nMin, nMax)

        # Count
        count = Counter(lst)

        # Showing
        mdl = QStandardItemModel()
        mdl.setHorizontalHeaderLabels([self.tr("Phrases"), self.tr("Frequency")])
        for i in count.most_common():
            word = QStandardItem(" ".join(i[0]))
            number = QStandardItem()
            number.setData(i[1], Qt.DisplayRole)
            if i[1] > 1:
                mdl.appendRow([word, number])

        self.tblPhrase.setModel(mdl)

    def analyzeWord(self):
        root = self.mw.mdlOutline.rootItem

        exclude = self.txtWordExclude.toPlainText().split(",")
        exclude = [e.strip().lower() for e in exclude]

        def listWords(item):
            txt = item.text()

            lst = re.findall(r"[\w']+", txt)
            for c in item.children():
                lst += listWords(c)

            return lst

        lst = listWords(root)
        lst2 = []

        # Cleaning
        for i in lst:
            if len(i) >= self.spnWordMin.value() and not i.lower() in exclude:
                lst2.append(i.lower())

        # Count
        count = Counter(lst2)

        # Showing
        mdl = QStandardItemModel()
        mdl.setHorizontalHeaderLabels([self.tr("Word"), self.tr("Frequency")])
        for i in count.most_common():
            word = QStandardItem(i[0])
            number = QStandardItem()
            number.setData(i[1], Qt.DisplayRole)
            mdl.appendRow([word, number])

        self.tblWord.setModel(mdl)

    def updateSettings(self):
        settings.frequencyAnalyzer["wordMin"] = self.spnWordMin.value()
        settings.frequencyAnalyzer["wordExclude"] = self.txtWordExclude.toPlainText()
        settings.frequencyAnalyzer["phraseMin"] = self.spnPhraseMin.value()
        settings.frequencyAnalyzer["phraseMax"] = self.spnPhraseMax.value()



#!/usr/bin/env python
# --!-- coding: utf8 --!--
import PyQt5
import os

features = {'qtwebkit': False, 'qtwebengine': False}

if 'QT_WEB' in os.environ:
    features[os.environ['QT_WEB']] = True
else:
    try:
        import PyQt5.QtWebKitWidgets
        features['qtwebkit'] = True
    except:
        features['qtwebkit'] = False

    try:
        import PyQt5.QtWebEngineWidgets
        features['qtwebengine'] = True
    except:
        features['qtwebengine'] = False

if features['qtwebkit']:
    from PyQt5.QtWebKitWidgets import QWebView
    print("Debug: Web rendering engine used: QWebView")
    webEngine = "QtWebKit"
    webView = QWebView
elif features['qtwebengine']:
    from PyQt5 import QtWebEngineWidgets
    print("Debug: Web rendering engine used: QWebEngineView")
    webEngine = "QtWebEngine"
    webView = QtWebEngineWidgets.QWebEngineView
else:
    from PyQt5.QtWidgets import QTextEdit
    print("Debug: Web rendering engine used: QTextEdit")
    webEngine = "QTextEdit"
    webView = QTextEdit



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'manuskript/ui/exporters/exportersManager_ui.ui'
#
# Created by: PyQt5 UI code generator 5.9
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_ExportersManager(object):
    def setupUi(self, ExportersManager):
        ExportersManager.setObjectName("ExportersManager")
        ExportersManager.resize(720, 548)
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout(ExportersManager)
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.lstExporters = QtWidgets.QListWidget(ExportersManager)
        self.lstExporters.setObjectName("lstExporters")
        item = QtWidgets.QListWidgetItem()
        self.lstExporters.addItem(item)
        self.horizontalLayout_8.addWidget(self.lstExporters)
        self.stack = QtWidgets.QWidget(ExportersManager)
        self.stack.setObjectName("stack")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.stack)
        self.verticalLayout_7.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.lblExporterName = QtWidgets.QLabel(self.stack)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lblExporterName.sizePolicy().hasHeightForWidth())
        self.lblExporterName.setSizePolicy(sizePolicy)
        self.lblExporterName.setText("{Exporter Name}")
        self.lblExporterName.setAlignment(QtCore.Qt.AlignCenter)
        self.lblExporterName.setObjectName("lblExporterName")
        self.verticalLayout_7.addWidget(self.lblExporterName)
        self.grpDescription = QtWidgets.QGroupBox(self.stack)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.grpDescription.setFont(font)
        self.grpDescription.setObjectName("grpDescription")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.grpDescription)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.lblExporterDescription = QtWidgets.QLabel(self.grpDescription)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.lblExporterDescription.setFont(font)
        self.lblExporterDescription.setText("{ExporterDescription}")
        self.lblExporterDescription.setWordWrap(True)
        self.lblExporterDescription.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.lblExporterDescription.setObjectName("lblExporterDescription")
        self.verticalLayout_5.addWidget(self.lblExporterDescription)
        self.verticalLayout_7.addWidget(self.grpDescription)
        self.grpExportTo = QtWidgets.QGroupBox(self.stack)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.grpExportTo.setFont(font)
        self.grpExportTo.setObjectName("grpExportTo")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.grpExportTo)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.lstExportTo = QtWidgets.QListWidget(self.grpExportTo)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lstExportTo.sizePolicy().hasHeightForWidth())
        self.lstExportTo.setSizePolicy(sizePolicy)
        self.lstExportTo.setObjectName("lstExportTo")
        self.horizontalLayout.addWidget(self.lstExportTo)
        self.frame = QtWidgets.QFrame(self.grpExportTo)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.lblExportToDescription = QtWidgets.QLabel(self.frame)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.lblExportToDescription.setFont(font)
        self.lblExportToDescription.setText("{FormatDescription}")
        self.lblExportToDescription.setWordWrap(True)
        self.lblExportToDescription.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.lblExportToDescription.setObjectName("lblExportToDescription")
        self.verticalLayout.addWidget(self.lblExportToDescription)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.horizontalLayout.addWidget(self.frame)
        self.verticalLayout_7.addWidget(self.grpExportTo)
        self.grpPath = QtWidgets.QGroupBox(self.stack)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.grpPath.setFont(font)
        self.grpPath.setObjectName("grpPath")
        self.formLayout = QtWidgets.QFormLayout(self.grpPath)
        self.formLayout.setObjectName("formLayout")
        self.label_2 = QtWidgets.QLabel(self.grpPath)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.lblStatus = QtWidgets.QLabel(self.grpPath)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.lblStatus.setFont(font)
        self.lblStatus.setText("{ExporterStatus}")
        self.lblStatus.setObjectName("lblStatus")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lblStatus)
        self.lblVersionName = QtWidgets.QLabel(self.grpPath)
        self.lblVersionName.setObjectName("lblVersionName")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.lblVersionName)
        self.lblVersion = QtWidgets.QLabel(self.grpPath)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.lblVersion.setFont(font)
        self.lblVersion.setText("{ExporterVersion}")
        self.lblVersion.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.lblVersion.setObjectName("lblVersion")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.lblVersion)
        self.label = QtWidgets.QLabel(self.grpPath)
        self.label.setObjectName("label")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.txtPath = QtWidgets.QLineEdit(self.grpPath)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.txtPath.setFont(font)
        self.txtPath.setObjectName("txtPath")
        self.horizontalLayout_2.addWidget(self.txtPath)
        self.btnSetPath = QtWidgets.QPushButton(self.grpPath)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.btnSetPath.setFont(font)
        self.btnSetPath.setObjectName("btnSetPath")
        self.horizontalLayout_2.addWidget(self.btnSetPath)
        self.formLayout.setLayout(3, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout_2)
        self.lblHelpText = QtWidgets.QLabel(self.grpPath)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.lblHelpText.setFont(font)
        self.lblHelpText.setWordWrap(True)
        self.lblHelpText.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.lblHelpText.setObjectName("lblHelpText")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.SpanningRole, self.lblHelpText)
        self.verticalLayout_7.addWidget(self.grpPath)
        self.horizontalLayout_8.addWidget(self.stack)

        self.retranslateUi(ExportersManager)
        QtCore.QMetaObject.connectSlotsByName(ExportersManager)

    def retranslateUi(self, ExportersManager):
        _translate = QtCore.QCoreApplication.translate
        ExportersManager.setWindowTitle(_translate("ExportersManager", "Manage Exporters"))
        __sortingEnabled = self.lstExporters.isSortingEnabled()
        self.lstExporters.setSortingEnabled(False)
        item = self.lstExporters.item(0)
        item.setText(_translate("ExportersManager", "Manuskript"))
        self.lstExporters.setSortingEnabled(__sortingEnabled)
        self.grpDescription.setTitle(_translate("ExportersManager", "Description"))
        self.grpExportTo.setTitle(_translate("ExportersManager", "Offers export to"))
        self.grpPath.setTitle(_translate("ExportersManager", "Status"))
        self.label_2.setText(_translate("ExportersManager", "Status:"))
        self.lblVersionName.setText(_translate("ExportersManager", "Version:"))
        self.label.setText(_translate("ExportersManager", "Path:"))
        self.btnSetPath.setText(_translate("ExportersManager", "..."))
        self.lblHelpText.setText(_translate("ExportersManager", "{HelpText}"))




#!/usr/bin/env python
# --!-- coding: utf8 --!--
from manuskript.exporter.pandoc.abstractPlainText import abstractPlainText


class abstractOutput(abstractPlainText):
    name = "SUBCLASSME"
    description = "SUBCLASSME"
    exportVarName = "SUBCLASSME"
    toFormat = "SUBCLASSME"
    icon = "SUBCLASSME"
    exportFilter = "SUBCLASSME"
    exportDefaultSuffix = ".SUBCLASSME"
    requires = {
        "Settings": True,
        "Preview": False,
    }

    def export(self, settingsWidget):
        filename = self.getExportFilename(settingsWidget)
        settingsWidget.writeSettings()
        if filename:
            content = self.output(settingsWidget, outputfile=filename)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re

from PyQt5.QtCore import QRegExp
from PyQt5.QtGui import QTextCursor


def MDFormatSelection(editor, style):
    """
    Formats the current selection of ``editor`` in the format given by ``style``, 
    style being:
        0: bold
        1: italic
        2: code
    """
    print("Formatting:", style, " (Unimplemented yet !)")
    # FIXME


#!/usr/bin/env python
# --!-- coding: utf8 --!--

# Lots of stuff from here comes from the excellent focuswriter.
import os
import re

from PyQt5.QtCore import QSettings, QRect, QSize, Qt, QPoint, QFile, QIODevice, QTextStream
from PyQt5.QtGui import QPixmap, QPainter, QColor, QBrush, QImage, QTextBlockFormat, QTextCharFormat, QFont, qGray
from PyQt5.QtWidgets import qApp, QFrame

from manuskript.functions import allPaths, appPath, findBackground, findFirstFile
from manuskript.ui.views.MDEditView import MDEditView

_thumbCache = {}

def loadThemeDatas(themeFile):
    settings = QSettings(themeFile, QSettings.IniFormat)
    _themeData = {}

    # Theme name
    _themeData["Name"] = getThemeName(themeFile)

    # Window Background
    loadThemeSetting(_themeData, settings, "Background/Color", "#000000")
    loadThemeSetting(_themeData, settings, "Background/ImageFile", "")
    loadThemeSetting(_themeData, settings, "Background/Type", 0)

    # Text Background
    loadThemeSetting(_themeData, settings, "Foreground/Color", "#ffffff")
    loadThemeSetting(_themeData, settings, "Foreground/Opacity", 50)
    loadThemeSetting(_themeData, settings, "Foreground/Margin", 40)
    loadThemeSetting(_themeData, settings, "Foreground/Padding", 10)
    loadThemeSetting(_themeData, settings, "Foreground/Position", 1)
    loadThemeSetting(_themeData, settings, "Foreground/Rounding", 5)
    loadThemeSetting(_themeData, settings, "Foreground/Width", 700)

    # Text Options
    loadThemeSetting(_themeData, settings, "Text/Color", "#ffffff")
    loadThemeSetting(_themeData, settings, "Text/Font", qApp.font().toString())
    loadThemeSetting(_themeData, settings, "Text/Misspelled", "#ff0000")

    # Paragraph Options
    loadThemeSetting(_themeData, settings, "Spacings/Alignment", 0)
    loadThemeSetting(_themeData, settings, "Spacings/IndentFirstLine", False)
    loadThemeSetting(_themeData, settings, "Spacings/LineSpacing", 100)
    loadThemeSetting(_themeData, settings, "Spacings/ParagraphAbove", 0)
    loadThemeSetting(_themeData, settings, "Spacings/ParagraphBelow", 0)
    loadThemeSetting(_themeData, settings, "Spacings/TabWidth", 48)

    return _themeData


def loadThemeSetting(datas, settings, key, default):
    """
    Loads data from ini file, using default value if the key is absent,
    and casting to the proper type based on default.
    """
    datas[key] = settings.value(key, default, type(default))

def getThemeName(theme):
    settings = QSettings(theme, QSettings.IniFormat)

    if settings.contains("Name"):
        return settings.value("Name")
    else:
        return os.path.splitext(os.path.split(theme)[1])[0]


def themeTextRect(themeDatas, screenRect):
    margin = themeDatas["Foreground/Margin"]
    x = 0
    y = margin
    width = min(themeDatas["Foreground/Width"], screenRect.width() - 2 * margin)
    height = screenRect.height() - 2 * margin

    if themeDatas["Foreground/Position"] == 0:  # Left
        x = margin
    elif themeDatas["Foreground/Position"] == 1:  # Center
        x = (screenRect.width() - width) / 2
    elif themeDatas["Foreground/Position"] == 2:  # Right
        x = screenRect.width() - margin - width
    elif themeDatas["Foreground/Position"] == 3:  # Stretched
        x = margin
        width = screenRect.width() - 2 * margin
    return QRect(x, y, width, height)


def createThemePreview(theme, screenRect, size=QSize(200, 120)):
    """
    Generates a QPixmap preview for given theme.

    Theme can be either a string containing the filename of the ini
    file with the theme settings, or it can be a dict with the settings.

    If theme is a filename, the result is cached.
    """

    # Checking whether theme is a string or dict
    if type(theme) == str and os.path.exists(theme):
        # Theme is the path to an ini file
        themeDatas = loadThemeDatas(theme)
        fromFile = True
    else:
        themeDatas = theme
        fromFile = False

    # Check if item is in cache
    if fromFile and theme in _thumbCache:
        if _thumbCache[theme][0] == themeDatas:
            return _thumbCache[theme][1]

    pixmap = generateTheme(themeDatas, screenRect)

    addThemePreviewText(pixmap, themeDatas, screenRect)

    px = QPixmap(pixmap).scaled(size, Qt.KeepAspectRatio)

    w = px.width() / 10
    h = px.height() / 10
    r = themeTextRect(themeDatas, screenRect)

    painter = QPainter(px)
    painter.drawPixmap(QRect(w, h, w * 4, h * 5), pixmap,
                       QRect(r.topLeft() - QPoint(w / 3, h / 3), QSize(w * 4, h * 5)))
    painter.setPen(Qt.white)
    painter.drawRect(QRect(w, h, w * 4, h * 5))
    painter.end()

    # If theme is a themefile, we keep it in cache
    if fromFile:
        _thumbCache[theme] = [themeDatas, px]

    return px


def findThemePath(themeName):
    p = findFirstFile(re.escape("{}.theme".format(themeName)), "resources/themes")
    if not p:
        return findFirstFile(r".*\.theme", "resources/themes")
    else:
        return p


def generateTheme(themeDatas, screenRect):
    # Window Background
    px = QPixmap(screenRect.size())
    px.fill(QColor(themeDatas["Background/Color"]))

    painter = QPainter(px)
    if themeDatas["Background/ImageFile"]:
        path = findBackground(themeDatas["Background/ImageFile"])
        _type = themeDatas["Background/Type"]
        if path and _type > 0:
            if _type == 1:  # Tiled
                painter.fillRect(screenRect, QBrush(QImage(path)))
            else:
                img = QImage(path)
                scaled = img.size()
                if _type == 3:  # Stretched
                    scaled.scale(screenRect.size(), Qt.IgnoreAspectRatio)
                elif _type == 4:  # Scaled
                    scaled.scale(screenRect.size(), Qt.KeepAspectRatio)
                elif _type == 5:  # Zoomed
                    scaled.scale(screenRect.size(), Qt.KeepAspectRatioByExpanding)

                painter.drawImage((screenRect.width() - scaled.width()) / 2,
                                  (screenRect.height() - scaled.height()) / 2, img.scaled(scaled))

    # Text Background
    textRect = themeTextRect(themeDatas, screenRect)

    painter.save()
    color = QColor(themeDatas["Foreground/Color"])
    color.setAlpha(themeDatas["Foreground/Opacity"] * 255 / 100)
    painter.setBrush(color)
    painter.setPen(Qt.NoPen)
    r = themeDatas["Foreground/Rounding"]
    painter.drawRoundedRect(textRect, r, r)
    painter.restore()

    painter.end()
    return px


def themeEditorGeometry(themeDatas, textRect):
    padding = themeDatas["Foreground/Padding"]
    x = textRect.x() + padding
    y = textRect.y() + padding + themeDatas["Spacings/ParagraphAbove"]
    width = textRect.width() - 2 * padding
    height = textRect.height() - 2 * padding - themeDatas["Spacings/ParagraphAbove"]
    return x, y, width, height


def getThemeBlockFormat(themeDatas):
    bf = QTextBlockFormat()
    bf.setAlignment(Qt.AlignLeft if themeDatas["Spacings/Alignment"] == 0 else
                    Qt.AlignCenter if themeDatas["Spacings/Alignment"] == 1 else
                    Qt.AlignRight if themeDatas["Spacings/Alignment"] == 2 else
                    Qt.AlignJustify)
    bf.setLineHeight(themeDatas["Spacings/LineSpacing"], QTextBlockFormat.ProportionalHeight)
    bf.setTextIndent(themeDatas["Spacings/TabWidth"] * 1 if themeDatas["Spacings/IndentFirstLine"] else 0)
    bf.setTopMargin(themeDatas["Spacings/ParagraphAbove"])
    bf.setBottomMargin(themeDatas["Spacings/ParagraphBelow"])
    return bf


def setThemeEditorDatas(editor, themeDatas, pixmap, screenRect):
    textRect = themeTextRect(themeDatas, screenRect)
    x, y, width, height = themeEditorGeometry(themeDatas, textRect)
    editor.setGeometry(x, y, width, height)

    # p = editor.palette()
    ##p.setBrush(QPalette.Base, QBrush(pixmap.copy(x, y, width, height)))
    # p.setBrush(QPalette.Base, QColor(Qt.transparent))
    # p.setColor(QPalette.Text, QColor(themeDatas["Text/Color"]))
    # p.setColor(QPalette.Highlight, QColor(themeDatas["Text/Color"]))
    # p.setColor(QPalette.HighlightedText, Qt.black if qGray(QColor(themeDatas["Text/Color"]).rgb()) > 127 else Qt.white)
    # editor.setPalette(p)

    editor.setAttribute(Qt.WA_NoSystemBackground, True)

    bf = getThemeBlockFormat(themeDatas)
    editor.setDefaultBlockFormat(bf)

    # b = editor.document().firstBlock()
    # cursor = editor.textCursor()
    # cursor.setBlockFormat(bf)
    # while b.isValid():
    # bf2 = b.blockFormat()
    # bf2.merge(bf)
    # cursor.setPosition(b.position())
    ##cursor.setPosition(b.position(), QTextCursor.KeepAnchor)
    # cursor.setBlockFormat(bf2)
    # b = b.next()

    editor.setTabStopWidth(themeDatas["Spacings/TabWidth"])
    editor.document().setIndentWidth(themeDatas["Spacings/TabWidth"])

    editor.highlighter.setMisspelledColor(QColor(themeDatas["Text/Misspelled"]))

    cf = QTextCharFormat()
    # f = QFont()
    # f.fromString(themeDatas["Text/Font"])
    # cf.setFont(f)
    editor.highlighter.setDefaultCharFormat(cf)
    f = QFont()
    f.fromString(themeDatas["Text/Font"])
    # editor.setFont(f)

    editor.setStyleSheet("""
        QTextEdit {{
            background: transparent;
            color: {foreground};
            font-family: {ff};
            font-size: {fs};
            selection-color: {sc};
            selection-background-color: {sbc};
        }}
        """.format(
            foreground=themeDatas["Text/Color"],
            ff=f.family(),
            fs="{}pt".format(str(f.pointSize())),
            sc="black" if qGray(QColor(themeDatas["Text/Color"]).rgb()) > 127 else "white",
            sbc=themeDatas["Text/Color"],
    )
    )

    editor._fromTheme = True
    editor._themeData = themeDatas
    editor.highlighter.updateColorScheme()

def addThemePreviewText(pixmap, themeDatas, screenRect):
    # Text
    previewText = MDEditView(highlighting=True)
    previewText.setFrameStyle(QFrame.NoFrame)
    previewText.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
    previewText.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
    f = QFile(appPath("resources/themes/preview.txt"))
    f.open(QIODevice.ReadOnly)
    previewText.setPlainText(QTextStream(f).readAll())

    setThemeEditorDatas(previewText, themeDatas, pixmap, screenRect)

    previewText.render(pixmap, previewText.pos())

    ## Text Background
    ##themeDatas["Foreground/Color"]
    ##themeDatas["Foreground/Opacity"]
    ##themeDatas["Foreground/Margin"]
    ##themeDatas["Foreground/Padding"]
    ##themeDatas["Foreground/Position"]
    ##themeDatas["Foreground/Rounding"]
    ##themeDatas["Foreground/Width"]

    ## Text Options
    ##themeDatas["Text/Color"]
    ##themeDatas["Text/Font"]
    # themeDatas["Text/Misspelled"]

    ## Paragraph Options
    ##themeDatas["Spacings/IndentFirstLine"]
    ##themeDatas["Spacings/LineSpacing"]
    ##themeDatas["Spacings/ParagraphAbove"]
    ##themeDatas["Spacings/ParagraphBelow"]
    ##themeDatas["Spacings/TabWidth"]



# --!-- coding: utf8 --!--

# Single source the package version
# https://packaging.python.org/guides/single-sourcing-package-version/

__version__ = "0.10.0"

def getVersion():
    return __version__



#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QToolBar, QDockWidget, QAction, QToolButton, QSizePolicy, QStylePainter, \
    QStyleOptionButton, QStyle

from manuskript.ui import style


class collapsibleDockWidgets(QToolBar):
    """
    QMainWindow "mixin" which provides auto-hiding support for dock widgets
    (not toolbars).
    """
    TRANSPOSED_AREA = {
        Qt.LeftDockWidgetArea: Qt.LeftToolBarArea,
        Qt.RightDockWidgetArea: Qt.RightToolBarArea,
        Qt.TopDockWidgetArea: Qt.TopToolBarArea,
        Qt.BottomDockWidgetArea: Qt.BottomToolBarArea,
    }

    def __init__(self, area, parent, name=""):
        QToolBar.__init__(self, parent)
        self._area = area
        if not name:
            name = self.tr("Dock Widgets Toolbar")
        self.setObjectName(name)
        self.setWindowTitle(name)

        self.setFloatable(False)
        self.setMovable(False)

        # self.setAllowedAreas(self.TRANSPOSED_AREA[self._area])
        self.parent().addToolBar(self.TRANSPOSED_AREA[self._area], self)

        self._dockToButtonAction = {}

        # Dock widgets
        for d in self._dockWidgets():
            b = verticalButton(self)
            b.setDefaultAction(d.toggleViewAction())
            # d.setStyleSheet("QDockWidget::title{background-color: red;}")
            # d.setTitleBarWidget(QLabel(d.windowTitle()))
            d.setStyleSheet(style.dockSS())
            a = self.addWidget(b)
            self._dockToButtonAction[d] = a

        self.addSeparator()

        # Other widgets
        self.otherWidgets = []
        self.currentGroup = None

        self.setStyleSheet(style.toolBarSS())
        self.layout().setContentsMargins(0,0,0,0)

    def _dockWidgets(self):
        mw = self.parent()
        for w in mw.findChildren(QDockWidget, None):
            yield w

    def addCustomWidget(self, text, widget, group=None, defaultVisibility=True):
        """
        Adds a custom widget to the toolbar.

        `text` is the name that will displayed on the button to switch visibility.
        `widget` is the widget to control from the toolbar.
        `group` is an integer (or any hashable) if the current widget should not
            be displayed all the time. Call `collapsibleDockWidgets.setCurrentGroup`
            to switch to that group and hide other widgets.
        `defaultVisibility` is the default visibility of the item when it is added.
            This allows for the widget to be added to `collapsibleDockWidgets` after
            they've been created but before they are shown, and yet specify their
            desired visibility. Otherwise it creates troubles, see #167 on github:
            https://github.com/olivierkes/manuskript/issues/167.
        """
        a = QAction(text, self)
        a.setCheckable(True)
        a.setChecked(defaultVisibility)
        a.toggled.connect(widget.setVisible)
        widget.setVisible(defaultVisibility)
        # widget.installEventFilter(self)
        b = verticalButton(self)
        b.setDefaultAction(a)
        #b.setChecked(widget.isVisible())
        a2 = self.addWidget(b)
        self.otherWidgets.append((b, a2, widget, group))

        # def eventFilter(self, widget, event):
        # if event.type() in [QEvent.Show, QEvent.Hide]:
        # for btn, action, w, grp in self.otherWidgets:
        # if w == widget:
        # btn.defaultAction().setChecked(event.type() == QEvent.Show)
        # return False

    def setCurrentGroup(self, group):
        self.currentGroup = group
        for btn, action, widget, grp in self.otherWidgets:
            if not grp == group or grp is None:
                action.setVisible(False)
            else:
                action.setVisible(True)

    def setDockVisibility(self, dock, val):
        dock.setVisible(val)
        self._dockToButtonAction[dock].setVisible(val)

    def saveState(self):
        """
        Saves and returns the state of the custom widgets. The visibility of the
        docks is not saved since it is included in `QMainWindow.saveState`.
        """
        state = []
        for btn, act, w, grp in self.otherWidgets:
            state.append(
                (grp, btn.text(), btn.isChecked())
            )
        return state

    def restoreState(self, state):
        """Restores the state of the custom widgets."""
        for group, title, status in state:
            for btn, act, widget, grp in self.otherWidgets:
                # Strip '&' from both title and btn.text() to improve matching because
                #   title contains "&" shortcut character whereas btn.text() does not.
                if group == grp and title.replace('&', '') == btn.text().replace('&', ''):
                    btn.setChecked(status)
                    btn.defaultAction().setChecked(status)
                    widget.setVisible(status)


class verticalButton(QToolButton):
    def __init__(self, parent):
        QToolButton.__init__(self, parent)
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)

        self.setStyleSheet(style.verticalToolButtonSS())

    def sizeHint(self):
        return QToolButton.sizeHint(self).transposed()

    def paintEvent(self, event):
        p = QStylePainter(self)
        p.rotate(90)
        p.translate(0, - self.width())
        opt = QStyleOptionButton()
        opt.initFrom(self)
        opt.text = self.text()
        if self.isChecked():
            opt.state |= QStyle.State_On
        s = opt.rect.size().transposed()
        opt.rect.setSize(s)
        p.drawControl(QStyle.CE_PushButton, opt)



#!/usr/bin/env python
# --!-- coding: utf8 --!--
import json
import os

from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtGui import QBrush, QColor, QIcon
from PyQt5.QtWidgets import QWidget, QFileDialog, QMessageBox, QStyle

from manuskript.functions import writablePath, appPath, openURL, statusMessage
from manuskript.ui.importers.importer_ui import Ui_importer
from manuskript.ui.importers.generalSettings import generalSettings
from manuskript.ui import style
from manuskript import importer
from manuskript.models import outlineModel, outlineItem
from manuskript.enums import Outline
from manuskript.exporter.pandoc import pandocExporter

class importerDialog(QWidget, Ui_importer):

    formatsIcon = {
        ".epub": "application-epub+zip",
        ".odt": "application-vnd.oasis.opendocument.text",
        ".docx": "application-vnd.openxmlformats-officedocument.wordprocessingml.document",
        ".md": "text-x-markdown",
        ".rst": "text-plain",
        ".tex": "text-x-tex",
        ".opml": "text-x-opml+xml",
        ".xml": "text-x-opml+xml",
        ".html": "text-html",
        }

    def __init__(self, parent=None, mw=None):
        QWidget.__init__(self, parent)
        self.setupUi(self)

        # Var
        self.mw = mw
        self.settingsWidget = None
        self.fileName = ""
        self.setStyleSheet(style.mainWindowSS())
        self.tree.setStyleSheet("QTreeView{background:transparent;}")
        self.editor.setStyleSheet("QWidget{background:transparent;}")
        self.editor.toggleSpellcheck(False)

        # Register importFormats:
        self.importers = importer.importers

        # Populate combo box with formats
        self.populateImportList()

        # Connections
        self.btnChoseFile.clicked.connect(self.selectFile)
        self.btnClearFileName.clicked.connect(self.setFileName)
        self.btnPreview.clicked.connect(self.preview)
        self.btnImport.clicked.connect(self.doImport)
        self.cmbImporters.currentTextChanged.connect(self.updateSettings)

        self.setFileName("")
        self.updateSettings()

    ############################################################################
    # Combobox / Formats
    ############################################################################

    def populateImportList(self):

        def addFormat(name, icon, identifier):
            # Identifier serves to distinguish 2 importers that would have the
            # same name.
            self.cmbImporters.addItem(QIcon.fromTheme(icon), name, identifier)

        def addHeader(name):
            self.cmbImporters.addItem(name, "header")
            self.cmbImporters.setItemData(self.cmbImporters.count() - 1, QBrush(QColor(style.highlightedTextDark)), Qt.ForegroundRole)
            self.cmbImporters.setItemData(self.cmbImporters.count() - 1, QBrush(QColor(style.highlightLight)), Qt.BackgroundRole)
            item = self.cmbImporters.model().item(self.cmbImporters.count() - 1)
            item.setFlags(Qt.ItemIsEnabled)

        lastEngine = ""

        for f in self.importers:
            # Header
            if f.engine != lastEngine:
                addHeader(f.engine)
                lastEngine = f.engine

            addFormat(f.name, f.icon, "{}:{}".format(f.engine, f.name))
            if not f.isValid():
                item = self.cmbImporters.model().item(self.cmbImporters.count() - 1)
                item.setFlags(Qt.NoItemFlags)

        if not pandocExporter().isValid():
            self.cmbImporters.addItem(
                self.style().standardIcon(QStyle.SP_MessageBoxWarning),
                "Install pandoc to import from much more formats",
                "::URL::http://pandoc.org/installing.html")

        self.cmbImporters.setCurrentIndex(1)

    def currentFormat(self):
        formatIdentifier = self.cmbImporters.currentData()

        if formatIdentifier == "header":
            return None

        F = [F for F in self.importers
               if formatIdentifier == "{}:{}".format(F.engine, F.name)][0]
        # We instantiate the class
        return F()

    ############################################################################
    # Import file
    ############################################################################

    def selectFile(self):
        """
        Called to select a file in the file system. Uses QFileDialog.
        """

        # We find the current selected format
        F = self._format

        options = QFileDialog.Options()
        if F.fileFormat == "<<folder>>":
            options = QFileDialog.ShowDirsOnly
            fileName = QFileDialog.getExistingDirectory(self, "Select import folder",
                                                        "", options=options)
        else:
            fileName, _ = QFileDialog.getOpenFileName(self, "Import from file", "",
                                                      F.fileFormat, options=options)
        self.setFileName(fileName)
        self.preview()

    def setFileName(self, fileName):
        """
        Updates Ui with given filename. Filename can be empty.
        """
        if fileName:
            self.fileName = fileName
            self.lblFileName.setText(os.path.basename(fileName))
            self.lblFileName.setToolTip(fileName)
            ext = os.path.splitext(fileName)[1]
            icon = None
            if ext and ext in self.formatsIcon:
                icon = QIcon.fromTheme(self.formatsIcon[ext])
            elif os.path.isdir(fileName):
                icon = QIcon.fromTheme("folder")

            if icon:
                self.lblIcon.setVisible(True)
                h = self.lblFileName.height()
                self.lblIcon.setPixmap(icon.pixmap(h, h))
            else:
                self.lblIcon.hide()

        else:
            self.fileName = None
            self.lblFileName.setText("")

        hasFile = True if fileName else False

        self.btnClearFileName.setVisible(hasFile)
        self.lblIcon.setVisible(hasFile)
        self.btnChoseFile.setVisible(not hasFile)
        self.btnPreview.setEnabled(hasFile)
        self.btnImport.setEnabled(hasFile)

    ############################################################################
    # UI
    ############################################################################

    def updateSettings(self):
        """
        When the current format change (through the combobox), we update the
        settings widget using the current format provided settings widget.
        """

        # We check if we have to open an URL
        data = self.cmbImporters.currentData()
        if data and data[:7] == "::URL::" and data[7:]:
            # FIXME: use functions.openURL after merge with feature/Exporters
            openURL(data[7:])
            return

        F = self.currentFormat()
        self._format = F

        # Checking if we have a valid importer (otherwise a header)
        if not F:
            self.grpSettings.setEnabled(False)
            self.grpPreview.setEnabled(False)
            return
        self.grpSettings.setEnabled(True)
        self.grpPreview.setEnabled(True)

        self.settingsWidget = generalSettings()
        #TODO: custom format widget to match exporter visuals?
        self.settingsWidget = F.settingsWidget(self.settingsWidget)

        # Set the settings widget in place
        self.setGroupWidget(self.grpSettings, self.settingsWidget)
        self.grpSettings.setMinimumWidth(200)

        # Clear file name
        self.setFileName("")

    def setGroupWidget(self, group, widget):
        """
        Sets the given widget as main widget for QGroupBox group.
        """

        # Removes every items from given layout.
        l = group.layout()
        while l.count():
            item = l.itemAt(0)
            l.removeItem(item)
            item.widget().deleteLater()

        l.addWidget(widget)
        widget.setParent(group)

    ############################################################################
    # Preview / Import
    ############################################################################

    def preview(self):

        if not self.fileName:
            return

        # Creating a temporary outlineModel
        previewModel = outlineModel(self)
        previewModel.loadFromXML(
            self.mw.mdlOutline.saveToXML(),
            fromString=True)

        # Inserting elements
        result = self.startImport(previewModel)

        if result:
            self.tree.setModel(previewModel)
            for i in range(1, previewModel.columnCount()):
                self.tree.hideColumn(i)
            self.tree.selectionModel().currentChanged.connect(self.editor.setCurrentModelIndex)
            self.previewSplitter.setStretchFactor(0, 10)
            self.previewSplitter.setStretchFactor(1, 40)

    def doImport(self):
        """
        Called by the Import button.
        """
        self.startImport(self.mw.mdlOutline)

        # Signal every views that important model changes have happened.
        self.mw.mdlOutline.layoutChanged.emit()

        # I'm getting segfault over this message sometimes...
        # Using status bar message instead...
        #QMessageBox.information(self, self.tr("Import status"),
                                #self.tr("Import Complete."))
        statusMessage("Import complete!", 5000)

        self.close()

    def startImport(self, outlineModel):
        """
        Where most of the magic happens.
        Is used by preview and by doImport (actual import).

        `outlineModel` is the model where the imported items are added.

        FIXME: Optimisation: when adding many outlineItems, outlineItem.updateWordCount
        is a bottleneck. It gets called a crazy number of time, and its not
        necessary.
        """

        items = []

        # We find the current selected format
        F = self._format

        # Parent item
        ID = self.settingsWidget.importUnderID()
        parentItem = outlineModel.getItemByID(ID)

        # Import in top-level folder?
        if self.settingsWidget.importInTopLevelFolder():
            parent = outlineItem(title=os.path.basename(self.fileName),
                                 parent=parentItem)
            parentItem = parent
            items.append(parent)

        # Calling the importer
        rItems = F.startImport(self.fileName,
                              parentItem,
                              self.settingsWidget)

        items.extend(rItems)

        # Do transformations
        items = self.doTransformations(items)

        return True

    def doTransformations(self, items):
        """
        Do general transformations.
        """

        # Trim long titles
        if self.settingsWidget.trimLongTitles():
            for item in items:
                if len(item.title()) > 32:
                    item.setData(Outline.title, item.title()[:32])

        # Split at
        if self.settingsWidget.splitScenes():
            for item in items:
                item.split(self.settingsWidget.splitScenes(), recursive=False)

        return items



#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import Qt, QRect
from PyQt5.QtGui import QColor, QPalette, QIcon, QFont, QFontMetrics
from PyQt5.QtWidgets import QStyledItemDelegate, qApp, QStyleOptionViewItem, QStyle

from manuskript import settings
from manuskript.enums import Outline
from manuskript.functions import mixColors, colorifyPixmap
from manuskript.functions import outlineItemColors
from manuskript.functions import toFloat
from manuskript.ui import style as S


class treeTitleDelegate(QStyledItemDelegate):
    """The main purpose of ``treeTitleDelegate`` is to paint outline items
    in the treeview with proper colors according to settings.
    """

    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)
        self._view = None

    def setView(self, view):
        self._view = view

    def paint(self, painter, option, index):

        item = index.internalPointer()
        colors = outlineItemColors(item)

        style = qApp.style()

        opt = QStyleOptionViewItem(option)
        self.initStyleOption(opt, index)

        iconRect = style.subElementRect(style.SE_ItemViewItemDecoration, opt)
        textRect = style.subElementRect(style.SE_ItemViewItemText, opt)

        # Background
        style.drawPrimitive(style.PE_PanelItemViewItem, opt, painter)

        if settings.viewSettings["Tree"]["Background"] != "Nothing" and not opt.state & QStyle.State_Selected:

            col = colors[settings.viewSettings["Tree"]["Background"]]

            if col != QColor(Qt.transparent):
                col2 = QColor(S.window)
                if opt.state & QStyle.State_Selected:
                    col2 = opt.palette.brush(QPalette.Normal, QPalette.Highlight).color()
                col = mixColors(col, col2, .2)

            painter.save()
            painter.setBrush(col)
            painter.setPen(Qt.NoPen)

            rect = opt.rect
            if self._view:
                r2 = self._view.visualRect(index)
                rect = self._view.viewport().rect()
                rect.setLeft(r2.left())
                rect.setTop(r2.top())
                rect.setBottom(r2.bottom())

            painter.drawRoundedRect(rect, 5, 5)
            painter.restore()

        # Icon
        mode = QIcon.Normal
        if not opt.state & QStyle.State_Enabled:
            mode = QIcon.Disabled
        elif opt.state & QStyle.State_Selected:
            mode = QIcon.Selected
        state = QIcon.On if opt.state & QStyle.State_Open else QIcon.Off
        icon = opt.icon.pixmap(iconRect.size(), mode=mode, state=state)
        if opt.icon and settings.viewSettings["Tree"]["Icon"] != "Nothing":
            color = colors[settings.viewSettings["Tree"]["Icon"]]
            colorifyPixmap(icon, color)
        opt.icon = QIcon(icon)
        opt.icon.paint(painter, iconRect, opt.decorationAlignment, mode, state)

        # Text
        if opt.text:
            painter.save()
            textColor = QColor(S.text)
            if option.state & QStyle.State_Selected:
                col = QColor(S.highlightedText)
                textColor = col
                painter.setPen(col)
            if settings.viewSettings["Tree"]["Text"] != "Nothing":
                col = colors[settings.viewSettings["Tree"]["Text"]]
                if col == Qt.transparent:
                    col = textColor
                # If text color is Compile and item is selected, we have
                # to change the color
                if settings.viewSettings["Outline"]["Text"] == "Compile" and \
                   not item.compile():
                    col = mixColors(textColor, QColor(S.window))
                painter.setPen(col)
            f = QFont(opt.font)
            painter.setFont(f)
            fm = QFontMetrics(f)
            elidedText = fm.elidedText(opt.text, Qt.ElideRight, textRect.width())
            painter.drawText(textRect, Qt.AlignLeft | Qt.AlignVCenter, elidedText)

            extraText = ""
            if item.isFolder() and settings.viewSettings["Tree"]["InfoFolder"] != "Nothing":
                if settings.viewSettings["Tree"]["InfoFolder"] == "Count":
                    extraText = item.childCount()
                    extraText = " [{}]".format(extraText)
                elif settings.viewSettings["Tree"]["InfoFolder"] == "WC":
                    extraText = item.wordCount()
                    extraText = " ({})".format(extraText)
                elif settings.viewSettings["Tree"]["InfoFolder"] == "Progress":
                    extraText = int(toFloat(item.data(Outline.goalPercentage)) * 100)
                    if extraText:
                        extraText = " ({}%)".format(extraText)
                elif settings.viewSettings["Tree"]["InfoFolder"] == "Summary":
                    extraText = item.data(Outline.summarySentence)
                    if extraText:
                        extraText = " - {}".format(extraText)

            if item.isText() and settings.viewSettings["Tree"]["InfoText"] != "Nothing":
                if settings.viewSettings["Tree"]["InfoText"] == "WC":
                    extraText = item.wordCount()
                    extraText = " ({})".format(extraText)
                elif settings.viewSettings["Tree"]["InfoText"] == "Progress":
                    extraText = int(toFloat(item.data(Outline.goalPercentage)) * 100)
                    if extraText:
                        extraText = " ({}%)".format(extraText)
                elif settings.viewSettings["Tree"]["InfoText"] == "Summary":
                    extraText = item.data(Outline.summarySentence)
                    if extraText:
                        extraText = " - {}".format(extraText)


            if extraText:
                r = QRect(textRect)
                r.setLeft(r.left() + fm.width(opt.text + " "))

                painter.save()
                f = painter.font()
                f.setWeight(QFont.Normal)
                painter.setFont(f)
                if option.state & QStyle.State_Selected:
                    col = QColor(S.highlightedTextLight)
                else:
                    col = QColor(S.textLight)
                painter.setPen(col)
                painter.drawText(r, Qt.AlignLeft | Qt.AlignVCenter, extraText)
                painter.restore()

            painter.restore()

            # QStyledItemDelegate.paint(self, painter, option, index)



#!/usr/bin/python
# -*- coding: utf8 -*-

from manuskript.ui.highlighters.basicHighlighter import BasicHighlighter
from manuskript.ui.highlighters.MMDHighlighter import MMDHighlighter

# Markdown highlighter
from manuskript.ui.highlighters.markdownEnums import MarkdownState
from manuskript.ui.highlighters.markdownEnums import MarkdownTokenType
from manuskript.ui.highlighters.markdownEnums import BlockquoteStyle
from manuskript.ui.highlighters.markdownTokenizer import MarkdownTokenizer
from manuskript.ui.highlighters.markdownHighlighter import MarkdownHighlighter



#!/usr/bin/env python
# --!-- coding: utf8 --!--

# The loadSave file calls the proper functions to load and save file
# trying to detect the proper file format if it comes from an older version
import os
import zipfile

import manuskript.load_save.version_0 as v0
import manuskript.load_save.version_1 as v1


def saveProject(version=None):

    # While debugging, we don't save the project
    # return

    if version == 0:
        return v0.saveProject()
    else:
        return v1.saveProject()


def clearSaveCache():
    v1.cache = {}


def loadProject(project):

    # Detect version
    isZip = False
    version = 0

    # Is it a zip?
    try:
        zf = zipfile.ZipFile(project)
        isZip = True
    except zipfile.BadZipFile:
        isZip = False

    # Does it have a VERSION in zip root?
    # Was used in transition between 0.2.0 and 0.3.0
    # So VERSION part can be deleted for manuskript 0.4.0
    if isZip and "VERSION" in zf.namelist():
        version = int(zf.read("VERSION"))

    # Does it have a MANUSKRIPT in zip root?
    elif isZip and "MANUSKRIPT" in zf.namelist():
        version = int(zf.read("MANUSKRIPT"))

    # Zip but no VERSION/MANUSKRIPT: oldest file format
    elif isZip:
        version = 0

    # Not a zip
    else:
        with open(project, "r", encoding="utf-8") as f:
            version = int(f.read())

    print("Loading:", project)
    print("Detected file format version: {}. Zip: {}.".format(version, isZip))

    if version == 0:
        v0.loadProject(project)
    else:
        v1.loadProject(project, zip=isZip)



#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QBrush, QColor
from PyQt5.QtWidgets import QComboBox

from manuskript.enums import Outline
from manuskript.functions import toInt
from manuskript.ui import style as S


class cmbOutlineCharacterChoser(QComboBox):
    def __init__(self, parent=None):
        QComboBox.__init__(self, parent)
        self.activated[int].connect(self.submit)
        self._column = Outline.POV
        self._index = None
        self._indexes = None
        self._updating = False
        self._various = False

    def setModels(self, mdlCharacter, mdlOutline):
        self.mdlCharacters = mdlCharacter
        self.mdlCharacters.dataChanged.connect(self.updateItems)
        self.mdlCharacters.rowsInserted.connect(self.updateItems)
        self.mdlCharacters.rowsRemoved.connect(self.updateItems)

        self.mdlOutline = mdlOutline
        self.mdlOutline.dataChanged.connect(self.update)
        self.updateItems()

    def updateItems(self):
        self.clear()
        self.addItem(QIcon.fromTheme("dialog-no"), self.tr("None"))

        l = [self.tr("Main"), self.tr("Secondary"), self.tr("Minor")]

        for importance in range(3):
            self.addItem(l[importance])
            self.setItemData(self.count() - 1, QBrush(QColor(S.highlightedTextDark)), Qt.ForegroundRole)
            self.setItemData(self.count() - 1, QBrush(QColor(S.highlightLight)), Qt.BackgroundRole)
            item = self.model().item(self.count() - 1)
            item.setFlags(Qt.ItemIsEnabled)
            for i in range(self.mdlCharacters.rowCount()):
                imp = toInt(self.mdlCharacters.importance(i))

                if not 2 - imp == importance:
                    continue

                self.addItem(self.mdlCharacters.icon(i), self.mdlCharacters.name(i), self.mdlCharacters.ID(i))
                self.setItemData(self.count() - 1, self.mdlCharacters.name(i), Qt.ToolTipRole)

        self._various = False

        if self._index or self._indexes:
            self.updateSelectedItem()

    def setCurrentModelIndex(self, index):
        self._indexes = None
        if index.column() != self._column:
            index = index.sibling(index.row(), self._column)
        self._index = index
        self.updateItems()

    def setCurrentModelIndexes(self, indexes):
        self._index = None

        idxes = []
        for i in indexes:
            if i.isValid():
                if i.column() != self._column:
                    i = i.sibling(i.row(), self._column)
                idxes.append(i)

        if idxes != self._indexes:
            self._indexes = idxes
            self.updateItems()

    def update(self, topLeft, bottomRight):

        if self._updating:
            # We are currently putting data in the model, so no updates
            return

        if self._index:
            if topLeft.row() <= self._index.row() <= bottomRight.row():
                self.updateSelectedItem()

        elif self._indexes:
            update = False
            for i in self._indexes:
                if topLeft.row() <= i.row() <= bottomRight.row():
                    update = True
            if update:
                self.updateSelectedItem()

    def getPOV(self, index):
        item = index.internalPointer()
        POV = item.data(self._column)
        return POV

    def selectPOV(self, POV):
        idx = self.findData(POV)
        if idx != -1:
            self.setCurrentIndex(idx)
        else:
            self.setCurrentIndex(0)

    def updateSelectedItem(self, idx1=None, idx2=None):

        if self._updating:
            return

        if self._index:
            POV = self.getPOV(self._index)
            self.selectPOV(POV)

        elif self._indexes:
            POVs = []
            same = True

            for i in self._indexes:
                POVs.append(self.getPOV(i))

            for POV in POVs[1:]:
                if POV != POVs[0]:
                    same = False
                    break

            if same:
                self._various = False
                self.selectPOV(POVs[0])

            else:
                if not self._various:
                    self.insertItem(0, self.tr("Various"))
                    f = self.font()
                    f.setItalic(True)
                    self.setItemData(0, f, Qt.FontRole)
                    self.setItemData(0, QBrush(Qt.darkGray), Qt.ForegroundRole)
                self._various = True
                self.setCurrentIndex(0)

        else:
            self.setCurrentIndex(0)

    def submit(self, idx):
        if self._index:
            self.mdlOutline.setData(self._index, self.currentData())

        elif self._indexes:
            if self._various and self.currentIndex() == 0:
                return

            self._updating = True
            for i in self._indexes:
                self.mdlOutline.setData(i, self.currentData())
            self._updating = False



#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtGui import QTextCharFormat, QFont
from PyQt5.QtWidgets import QPlainTextEdit, QGroupBox, qApp, QVBoxLayout, QCheckBox

from manuskript.exporter.manuskript.plainText import plainText
from manuskript.functions import mainWindow
from manuskript.ui.highlighters import MMDHighlighter
from manuskript.ui.exporters.manuskript.plainTextSettings import exporterSettings


class markdown(plainText):
    name = "Markdown"
    description = qApp.translate("Export", """Just like plain text, excepts adds markdown titles.
                          Presupposes that texts are formatted in markdown.""")

    exportVarName = "lastManuskriptMarkdown"
    exportFilter = "Markdown files (*.md);; Any files (*)"
    exportDefaultSuffix = ".md"
    icon = "text-x-markdown"

    def settingsWidget(self):
        w = markdownSettings(self)
        w.loadSettings()
        return w

    def preparesTextEditViewMarkdown(self, view, settings):
        if settings["Preview"]["MarkdownHighlighter"]:
            self.highlighter = MMDHighlighter(view)
        else:
            self.highlighter = None

    def preview(self, settingsWidget, previewWidget):
        settings = settingsWidget.getSettings()

        # Save settings
        settingsWidget.writeSettings()

        # Prepares text edit
        self.preparesTextEditViewMarkdown(previewWidget, settingsWidget.settings)
        self.preparesTextEditView(previewWidget, settings["Preview"]["PreviewFont"])

        previewWidget.setPlainText(self.output(settingsWidget))

    def processTitle(self, text, level, settings):
        return "{} {}\n".format(
            "#" * (level + 1),
            text
        )


class markdownSettings(exporterSettings):
    def __init__(self, _format, parent=None):
        exporterSettings.__init__(self, _format, parent)

        # Adds markdown syntax highlighter setting
        w = self.toolBox.widget(self.toolBox.count() - 1)
        self.grpMarkdown = QGroupBox(self.tr("Markdown"))
        self.grpMarkdown.setLayout(QVBoxLayout())
        self.chkMarkdownHighlighter = QCheckBox(qApp.translate("Export", "Preview with highlighter."))
        self.grpMarkdown.layout().addWidget(self.chkMarkdownHighlighter)

        w.layout().insertWidget(w.layout().count() - 1, self.grpMarkdown)

    def updateFromSettings(self):
        exporterSettings.updateFromSettings(self)

        s = self.settings["Preview"]
        val = s.get("MarkdownHighlighter", False)
        self.chkMarkdownHighlighter.setChecked(val)

    def getSettings(self):
        self.settings = exporterSettings.getSettings(self)
        self.settings["Preview"]["MarkdownHighlighter"] = self.chkMarkdownHighlighter.isChecked()

        return self.settings



# --!-- coding: utf8 --!--

from PyQt5.Qt import PYQT_VERSION_STR
from PyQt5.QtCore import QT_VERSION_STR
from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtWidgets import QWidget
from platform import python_version

from manuskript.functions import appPath
from manuskript.ui.about_ui import Ui_about
from manuskript.version import getVersion

class aboutDialog(QWidget, Ui_about):
    def __init__(self, parent=None, mw=None):
        QWidget.__init__(self, parent)
        self.setupUi(self)
        self.populateFields()
        self.buttonBox.accepted.connect(self.accept)

    def populateFields(self):
        # Fill in all the fields in the About dialog
        iconPic = appPath("icons/Manuskript/icon-64px.png")
        self.setWindowIcon(QIcon(iconPic))

        logoPic = QPixmap(appPath("icons/Manuskript/logo-400x104.png"))
        self.labelLogo.setPixmap(logoPic)

        self.labelManuskriptVersion.setText(
              "<b>" + self.tr("Version") + " " + getVersion() + "</b><br>"
            + "&nbsp;"*5 + """<a href="http://www.theologeek.ch/manuskript/">
                                http://www.theologeek.ch/manuskript/
                               </a><br>"""
            + "&nbsp;"*5 + "Copyright © 2015-2019 Olivier Keshavjee<br>"
            + "&nbsp;"*5 + """<a href="https://www.gnu.org/licenses/gpl-3.0.en.html">
                                GNU General Public License Version 3
                            </a><br>"""
            )

        self.labelManuskriptVersion.setOpenExternalLinks(True)

        self.labelSoftwareVersion.setText(
              "<b>" + self.tr("Software Versions in Use:") + "</b><br>"
            + "&nbsp;"*5 + "Python " + python_version() + "<br>"
            + "&nbsp;"*5 + "PyQt " + PYQT_VERSION_STR + "<br>"
            + "&nbsp;"*5 + "Qt " + QT_VERSION_STR
            )
        #self.labelPythonVersion.setText()
        #self.labelPyQtVersion.setText()
        #self.labelQtVersion.setText()

    def accept(self):
        self.close()



#!/usr/bin/env python
# --!-- coding: utf8 --!--

import re

from PyQt5.QtCore import QRegExp, Qt, QTimer, QRect, QPoint
from PyQt5.QtGui import QTextCursor
from PyQt5.QtWidgets import qApp, QToolTip

from manuskript.ui.views.textEditView import textEditView
from manuskript.ui.highlighters import MarkdownHighlighter
from manuskript import settings
from manuskript.ui.highlighters.markdownEnums import MarkdownState as MS
from manuskript.ui.highlighters.markdownTokenizer import MarkdownTokenizer as MT
from manuskript import functions as F


class MDEditView(textEditView):

    blockquoteRegex = QRegExp("^ {0,3}(>\\s*)+")
    listRegex = QRegExp(r"^(\s*)([+*-]|([0-9a-z])+([.\)]))(\s+)")
    inlineLinkRegex = QRegExp("\\[([^\n]+)\\]\\(([^\n]+)\\)")
    imageRegex = QRegExp("!\\[([^\n]*)\\]\\(([^\n]+)\\)")
    automaticLinkRegex = QRegExp("(<([a-zA-Z]+\\:[^\n]+)>)|(<([^\n]+@[^\n]+)>)")

    def __init__(self, parent=None, index=None, html=None, spellcheck=None,
                 highlighting=False, dict="", autoResize=False):
        textEditView.__init__(self, parent, index, html, spellcheck,
                              highlighting=True, dict=dict,
                              autoResize=autoResize)

        # Highlighter
        self._textFormat = "md"
        self._highlighterClass = MarkdownHighlighter
        self._noFocusMode = False
        self._lastCursorPosition = None

        if index:
            # We have to setup things anew, for the highlighter notably
            self.setCurrentModelIndex(index)

        self.cursorPositionChanged.connect(self.cursorPositionHasChanged)
        self.verticalScrollBar().rangeChanged.connect(
            self.scrollBarRangeChanged)

        # Clickable things
        self.clickRects = []
        self.textChanged.connect(self.getClickRects)
        self.document().documentLayoutChanged.connect(self.getClickRects)
        self.setMouseTracking(True)

    ###########################################################################
    # KEYPRESS
    ###########################################################################

    def keyPressEvent(self, event):
        k = event.key()
        m = event.modifiers()
        cursor = self.textCursor()

        # RETURN
        if k == Qt.Key_Return:
            if not cursor.hasSelection():
                if m & Qt.ShiftModifier:
                    # Insert Markdown-style line break
                    cursor.insertText("  ")

                if m & Qt.ControlModifier:
                    cursor.insertText("\n")
                else:
                    self.handleCarriageReturn()
            else:
                textEditView.keyPressEvent(self, event)

        # TAB
        elif k == Qt.Key_Tab:
            #self.indentText()
            # FIXME
            textEditView.keyPressEvent(self, event)
        elif k == Qt.Key_Backtab:
            #self.unindentText()
            # FIXME
            textEditView.keyPressEvent(self, event)

        else:
            textEditView.keyPressEvent(self, event)

    # Thanks to GhostWriter, mainly
    def handleCarriageReturn(self):
        autoInsertText = "";
        cursor = self.textCursor()
        endList = False
        moveBack = False
        text = cursor.block().text()

        if cursor.positionInBlock() < cursor.block().length() - 1:
            autoInsertText = self.getPriorIndentation()
            if cursor.positionInBlock() < len(autoInsertText):
                autoInsertText = autoInsertText[:cursor.positionInBlock()]

        else:
            s = cursor.block().userState()

            if s in [MS.MarkdownStateNumberedList,
                     MS.MarkdownStateBulletPointList]:
                self.listRegex.indexIn(text)
                g = self.listRegex.capturedTexts()
                    # 0 = "   a. " or "  * "
                    # 1 = "   "       "  "
                    # 2 =    "a."       "*"
                    # 3 =    "a"          ""
                    # 4 =     "."         ""
                    # 5 =      " "        " "

                # If the line of text is an empty list item, end the list.
                if len(g[0].strip()) == len(text.strip()):
                    endList = True

                # Else increment the list number
                elif g[3]:  # Numbered list
                    try: # digit
                        i = int(g[3])+1

                    except: # letter
                        i = chr(ord(g[3])+1)

                    autoInsertText = "{}{}{}{}".format(
                            g[1], i, g[4], g[5])

                else:  # Bullet list
                    autoInsertText = g[0]

                if text[-2:] == "  ":
                    autoInsertText = " " * len(autoInsertText)

            elif s == MS.MarkdownStateBlockquote:
                self.blockquoteRegex.indexIn(text)
                g = self.blockquoteRegex.capturedTexts()
                autoInsertText = g[0]

            elif s in [MS.MarkdownStateInGithubCodeFence,
                       MS.MarkdownStateInPandocCodeFence] and \
                 cursor.block().previous().userState() != s:
                autoInsertText = "\n" + text
                moveBack = True

            else:
                autoInsertText = self.getPriorIndentation()

        # Clear the list
        if endList:
            autoInsertText = self.getPriorIndentation()
            cursor.movePosition(QTextCursor.StartOfBlock)
            cursor.movePosition(QTextCursor.EndOfBlock, QTextCursor.KeepAnchor)
            cursor.insertText(autoInsertText)
            autoInsertText = ""

        # Finally, we insert
        cursor.insertText("\n" + autoInsertText)
        if moveBack:
            cursor.movePosition(QTextCursor.PreviousBlock)
            self.setTextCursor(cursor)

        self.ensureCursorVisible()

    def getPriorIndentation(self):
        text = self.textCursor().block().text()
        l = len(text) - len(text.lstrip())
        return text[:l]

    def getPriorMarkdownBlockItemStart(self, itemRegex):
        text = self.textCursor().block().text()
        if itemRegex.indexIn(text) >= 0:
            return text[itemRegex.matchedLength():]

        return ""

    ###########################################################################
    # TypeWriterScrolling
    ###########################################################################

    def setCurrentModelIndex(self, index):
        textEditView.setCurrentModelIndex(self, index)
        self.centerCursor()

    def cursorPositionHasChanged(self):
        self.centerCursor()
        # Focus mode
        if self.highlighter and settings.textEditor["focusMode"]:
            if self._lastCursorPosition:
                block = self.document().findBlock(self._lastCursorPosition)
                self.highlighter.rehighlightBlock(block)
            self._lastCursorPosition = self.textCursor().position()
            block = self.document().findBlock(self._lastCursorPosition)
            self.highlighter.rehighlightBlock(block)

    def centerCursor(self, force=False):
        cursor = self.cursorRect()
        scrollbar = self.verticalScrollBar()
        viewport = self.viewport().rect()
        if (force or settings.textEditor["alwaysCenter"]
                or cursor.bottom() >= viewport.bottom()
                or cursor.top() <= viewport.top()):
            offset = viewport.center() - cursor.center()
            scrollbar.setValue(scrollbar.value() - offset.y())

    def scrollBarRangeChanged(self, min, max):
        """
        Adds viewport height to scrollbar max so that we can center cursor
        on screen.
        """
        if settings.textEditor["alwaysCenter"]:
            self.verticalScrollBar().blockSignals(True)
            self.verticalScrollBar().setMaximum(max + self.viewport().height())
            self.verticalScrollBar().blockSignals(False)

    ###########################################################################
    # FORMATTING
    ###########################################################################

    def bold(self): self.insertFormattingMarkup("**")
    def italic(self): self.insertFormattingMarkup("*")
    def strike(self): self.insertFormattingMarkup("~~")
    def verbatim(self): self.insertFormattingMarkup("`")
    def superscript(self): self.insertFormattingMarkup("^")
    def subscript(self): self.insertFormattingMarkup("~")
    def blockquote(self): self.lineFormattingMarkup("> ")
    def orderedList(self): self.lineFormattingMarkup(" 1. ")
    def unorderedList(self): self.lineFormattingMarkup("  - ")

    def selectWord(self, cursor):
        if cursor.selectedText():
            return
        end = cursor.selectionEnd()
        cursor.movePosition(QTextCursor.StartOfWord)
        cursor.setPosition(end, QTextCursor.KeepAnchor)
        cursor.movePosition(QTextCursor.EndOfWord, QTextCursor.KeepAnchor)

    def selectBlock(self, cursor):
        cursor.movePosition(QTextCursor.StartOfBlock)
        cursor.movePosition(QTextCursor.EndOfBlock, QTextCursor.KeepAnchor)

    def comment(self):
        cursor = self.textCursor()

        # Select beginning and end of words
        self.selectWord(cursor)

        if cursor.hasSelection():
            text = cursor.selectedText()
            cursor.insertText("<!-- " + text + " -->")
        else:
            cursor.insertText("<!--  -->")
            cursor.movePosition(QTextCursor.PreviousCharacter,
                                QTextCursor.MoveAnchor, 4)
            self.setTextCursor(cursor)

    def commentLine(self):
        cursor = self.textCursor()

        start = cursor.selectionStart()
        end = cursor.selectionEnd()
        block = self.document().findBlock(start)
        block2 = self.document().findBlock(end)

        if True:
            # Method 1
            cursor.beginEditBlock()
            while block.isValid():
                self.commentBlock(block)
                if block == block2: break
                block = block.next()
            cursor.endEditBlock()

        else:
            # Method 2
            cursor.beginEditBlock()
            cursor.setPosition(block.position())
            cursor.insertText("<!--\n")
            cursor.setPosition(block2.position() + block2.length() - 1)
            cursor.insertText("\n-->")
            cursor.endEditBlock()

    def commentBlock(self, block):
        cursor = QTextCursor(block)
        text = block.text()
        if text[:5] == "<!-- " and \
           text[-4:] == " -->":
            text2 = text[5:-4]
        else:
            text2 = "<!-- " + text + " -->"
        self.selectBlock(cursor)
        cursor.insertText(text2)

    def lineFormattingMarkup(self, markup):
        """
        Adds `markup` at the beginning of block.
        """
        cursor = self.textCursor()
        cursor.movePosition(cursor.StartOfBlock)
        cursor.insertText(markup)

    def insertFormattingMarkup(self, markup):
        cursor = self.textCursor()

        # Select beginning and end of words
        self.selectWord(cursor)

        if cursor.hasSelection():
            start = cursor.selectionStart()
            end = cursor.selectionEnd() + len(markup)
            cursor.beginEditBlock()
            cursor.setPosition(start)
            cursor.insertText(markup)
            cursor.setPosition(end)
            cursor.insertText(markup)
            cursor.endEditBlock()
            cursor.movePosition(QTextCursor.PreviousCharacter,
                                QTextCursor.KeepAnchor, len(markup))
            #self.setTextCursor(cursor)

        else:
            # Insert markup twice (for opening and closing around the cursor),
            # and then move the cursor to be between the pair.
            cursor.beginEditBlock()
            cursor.insertText(markup)
            cursor.insertText(markup)
            cursor.movePosition(QTextCursor.PreviousCharacter,
                                QTextCursor.MoveAnchor, len(markup))
            cursor.endEditBlock()
            self.setTextCursor(cursor)

    def clearFormat(self):
        cursor = self.textCursor()
        text = cursor.selectedText()
        if not text:
            self.selectBlock(cursor)
            text = cursor.selectedText()
        text = self.clearedFormat(text)
        cursor.insertText(text)

    def clearedFormat(self, text):
        # FIXME: clear also block formats
        for reg, rep, flags in [
            (r"\*\*(.*?)\*\*", "\\1", None), # bold
            ("__(.*?)__", "\\1", None), # bold
            (r"\*(.*?)\*", "\\1", None), # emphasis
            ("_(.*?)_", "\\1", None), # emphasis
            ("`(.*?)`", "\\1", None), # verbatim
            ("~~(.*?)~~", "\\1", None), # strike
            (r"\^(.*?)\^", "\\1", None), # superscript
            ("~(.*?)~", "\\1", None), # subscript
            (r"<!--\s*(.*?)\s*-->", "\\1", re.S), # comments

            # LINES OR BLOCKS
            (r"^#*\s*(.+?)\s*", "\\1", re.M), # ATX
            (r"^[=-]*$", "", re.M), # Setext
            (r"^`*$", "", re.M), # Code block fenced
            (r"^\s*[-+*]\s*(.*?)\s*$", "\\1", re.M), # Bullet List
            (r"^\s*[0-9a-z](\.|\))\s*(.*?)\s*$", "\\2", re.M), # Bullet List
            (r"\s*[>\s]*(.*?)\s*$", "\\1", re.M), # Code block and blockquote

            ]:
            text = re.sub(reg, rep, text, flags if flags else 0)
        return text

    def clearedFormatForStats(self, text):
        # Remove stuff that musn't be counted
        # FIXME: clear also block formats
        for reg, rep, flags in [
            ("<!--.*-->", "", re.S), # comments
            ]:
            text = re.sub(reg, rep, text, flags if flags else 0)
        return text

    def titleSetext(self, level):
        cursor = self.textCursor()

        cursor.beginEditBlock()
        # Is it already a Setext header?
        if cursor.block().userState() in [
                MS.MarkdownStateSetextHeading1Line2,
                MS.MarkdownStateSetextHeading2Line2]:
            cursor.movePosition(QTextCursor.PreviousBlock)

        text = cursor.block().text()

        if cursor.block().userState() in [
                MS.MarkdownStateSetextHeading1Line1,
                MS.MarkdownStateSetextHeading2Line1]:
            # Need to remove line below
            c = QTextCursor(cursor.block().next())
            self.selectBlock(c)
            c.insertText("")

        char = "=" if level == 1 else "-"
        text = re.sub(r"^#*\s*(.*)\s*#*", "\\1", text)  # Removes #
        sub = char * len(text)
        text = text + "\n" + sub

        self.selectBlock(cursor)
        cursor.insertText(text)
        cursor.endEditBlock()

    def titleATX(self, level):
        cursor = self.textCursor()
        text = cursor.block().text()

        # Are we in a Setext Header?
        if cursor.block().userState() in [
                MS.MarkdownStateSetextHeading1Line1,
                MS.MarkdownStateSetextHeading2Line1]:
            # Need to remove line below
            cursor.beginEditBlock()
            c = QTextCursor(cursor.block().next())
            self.selectBlock(c)
            c.insertText("")

            self.selectBlock(cursor)
            cursor.insertText(text)
            cursor.endEditBlock()
            return

        elif cursor.block().userState() in [
                MS.MarkdownStateSetextHeading1Line2,
                MS.MarkdownStateSetextHeading2Line2]:
            cursor.movePosition(QTextCursor.PreviousBlock)
            self.setTextCursor(cursor)
            self.titleATX(level)
            return

        m = re.match(r"^(#+)(\s*)(.+)", text)
        if m:
            pre = m.group(1)
            space = m.group(2)
            txt = m.group(3)

            if len(pre) == level:
                # Remove title
                text = txt
            else:
                text = "#" * level + space + txt

        else:
            text = "#" * level + " " + text

        self.selectBlock(cursor)
        cursor.insertText(text)

    ###########################################################################
    # CLICKABLE THINKS
    ###########################################################################

    def resizeEvent(self, event):
        textEditView.resizeEvent(self, event)
        self.getClickRects()

    def scrollContentsBy(self, dx, dy):
        textEditView.scrollContentsBy(self, dx, dy)
        self.getClickRects()

    def getClickRects(self):
        """
        Parses the whole texte to catch clickable things: links and images.
        Stores the result so that it can be used elsewhere.
        """
        cursor = self.textCursor()
        refs = []
        text = self.toPlainText()
        for rx in [
                self.imageRegex,
                self.automaticLinkRegex,
                self.inlineLinkRegex,
            ]:
            pos = 0
            while rx.indexIn(text, pos) != -1:
                cursor.setPosition(rx.pos())
                r1 = self.cursorRect(cursor)
                pos = rx.pos() + rx.matchedLength()
                cursor.setPosition(pos)
                r2 = self.cursorRect(cursor)
                if r1.top() == r2.top():
                    ct = ClickThing(
                            QRect(r1.topLeft(), r2.bottomRight()),
                            rx,
                            rx.capturedTexts())
                    refs.append(ct)
                else:
                    r1.setRight(self.viewport().geometry().right())
                    refs.append(ClickThing(r1, rx, rx.capturedTexts()))
                    r2.setLeft(self.viewport().geometry().left())
                    refs.append(ClickThing(r2, rx, rx.capturedTexts()))
                    # We check for middle lines
                    cursor.setPosition(rx.pos())
                    cursor.movePosition(cursor.Down)
                    while self.cursorRect(cursor).top() != r2.top():
                        r3 = self.cursorRect(cursor)
                        r3.setLeft(self.viewport().geometry().left())
                        r3.setRight(self.viewport().geometry().right())
                        refs.append(ClickThing(r3, rx, rx.capturedTexts()))
                        cursor.movePosition(cursor.Down)

        self.clickRects = refs

    def mouseMoveEvent(self, event):
        """
        When mouse moves, we show tooltip when appropriate.
        """
        textEditView.mouseMoveEvent(self, event)

        onRect = [r for r in self.clickRects if r.rect.contains(event.pos())]

        if not onRect:
            qApp.restoreOverrideCursor()
            QToolTip.hideText()
            return

        ct = onRect[0]
        if not qApp.overrideCursor():
            qApp.setOverrideCursor(Qt.PointingHandCursor)

        if ct.regex == self.automaticLinkRegex:
            tooltip = ct.texts[2] or ct.texts[4]

        elif ct.regex == self.imageRegex:
            tt = ("<p><b>" + ct.texts[1] + "</b></p>"
                  +"<p><img src='data:image/png;base64,{}'></p>")
            tooltip = None
            pos = event.pos() + QPoint(0, ct.rect.height())
            ImageTooltip.fromUrl(ct.texts[2], pos, self)

        elif ct.regex == self.inlineLinkRegex:
            tooltip = ct.texts[1] or ct.texts[2]

        if tooltip:
            tooltip = self.tr("{} (CTRL+Click to open)").format(tooltip)
            QToolTip.showText(self.mapToGlobal(event.pos()), tooltip)

    def mouseReleaseEvent(self, event):
        textEditView.mouseReleaseEvent(self, event)
        onRect = [r for r in self.clickRects if r.rect.contains(event.pos())]
        if onRect and event.modifiers() & Qt.ControlModifier:
            ct = onRect[0]

            if ct.regex == self.automaticLinkRegex:
                url = ct.texts[2] or ct.texts[4]
            elif ct.regex == self.imageRegex:
                url = ct.texts[2]
            elif ct.regex == self.inlineLinkRegex:
                url = ct.texts[2]

            F.openURL(url)
            qApp.restoreOverrideCursor()

    # def paintEvent(self, event):
    #     """
    #     Only useful for debugging: shows which rects are detected for
    #     clickable things.
    #     """
    #     textEditView.paintEvent(self, event)
    #
    #     # Debug: paint rects
    #     from PyQt5.QtGui import QPainter
    #     painter = QPainter(self.viewport())
    #     painter.setPen(Qt.gray)
    #     for r in self.clickRects:
    #         painter.drawRect(r.rect)

    def doTooltip(self, pos, message):
        QToolTip.showText(self.mapToGlobal(pos), message)

class ClickThing:
    """
    A simple class to remember QRect associated with clickable stuff.
    """
    def __init__(self, rect, regex, texts):
        self.rect = rect
        self.regex = regex
        self.texts = texts

from PyQt5.QtNetwork import QNetworkRequest, QNetworkAccessManager, QNetworkReply
from PyQt5.QtCore import QIODevice, QUrl, QBuffer
from PyQt5.QtGui import QPixmap

class ImageTooltip:
    """
    This class handles the retrieving and caching of images in order to display these in tooltips.
    """

    cache = {}
    manager = QNetworkAccessManager()
    processing = {}

    supportedSchemes = ("", "file", "http", "https")

    def fromUrl(url, pos, editor):
        """
        Shows the image tooltip for the given url if available, or requests it for future use.
        """
        ImageTooltip.editor = editor

        if ImageTooltip.showTooltip(url, pos):
            return # the url already exists in the cache

        try:
            ImageTooltip.manager.finished.connect(ImageTooltip.finished, F.AUC)
        except:
            pass # already connected

        qurl = QUrl.fromUserInput(url)
        if (qurl == QUrl()):
            ImageTooltip.cache[url] = (False, ImageTooltip.manager.tr("The image path or URL is incomplete or malformed."))
            ImageTooltip.showTooltip(url, pos)
            return # empty QUrl means it failed completely
        elif (qurl.scheme() not in ImageTooltip.supportedSchemes):
            # QUrl.fromUserInput() can occasionally deduce an incorrect scheme,
            # which produces an error message regarding an unknown scheme. (Yay!)
            # But it also breaks all possible methods to try and associate the
            # reply with the original request in finished(), since reply.request()
            # is completely and utterly butchered for all tracking needs. :'(
            # (The QNetworkRequest, .url() and .originatingObject() can all change.)

            # Test case (Linux): ![image](C:\test_root.jpg)
            ImageTooltip.cache[url] = (False, ImageTooltip.manager.tr("The protocol \"{}\" is not supported.").format(qurl.scheme()))
            ImageTooltip.showTooltip(url, pos)
            return # no more request/reply chaos, please!
        elif (qurl in ImageTooltip.processing):
            return # one download is more than enough

        # Request the image for later processing.
        request = QNetworkRequest(qurl)
        ImageTooltip.processing[qurl] = (pos, url)
        reply = ImageTooltip.manager.get(request)

        # On Linux the finished() signal is not triggered when the url resembles
        # 'file://X:/...'. But because it completes instantly, we can manually
        # trigger the code to keep our processing dictionary neat & clean.
        if reply.error() == 302:  # QNetworkReply.ProtocolInvalidOperationError
            ImageTooltip.finished(reply)

    def finished(reply):
        """
        After retrieving an image, we add it to the cache.
        """
        cache = ImageTooltip.cache
        url_key = reply.request().url()
        pos, url = None, None

        if url_key in ImageTooltip.processing:
            # Obtain the information associated with this request.
            pos, url = ImageTooltip.processing[url_key]
            del ImageTooltip.processing[url_key]
        elif len(ImageTooltip.processing) == 0:
            # We are not processing anything. Maybe it is a spurious signal,
            # or maybe the 'reply.error() == 302' workaround in fromUrl() has
            # been fixed in Qt. Whatever the reason, we can assume this request
            # has already been handled, and needs no more work from us.
            return
        else:
            # Somehow we lost track. Log what we can to hopefully figure it out.
            print("Warning: unable to match fetched data for tooltip to original request.")
            print("- Completed request:", url_key)
            print("- Status upon finishing:", reply.error(), reply.errorString())
            print("- Currently processing:", ImageTooltip.processing)
            return

        # Update cache with retrieved data.
        if reply.error() != QNetworkReply.NoError:
            cache[url] = (False, reply.errorString())
        else:
            px = QPixmap()
            px.loadFromData(reply.readAll())
            px = px.scaled(800, 600, Qt.KeepAspectRatio)
            cache[url] = (True, px)

        ImageTooltip.showTooltip(url, pos)

    def showTooltip(url, pos):
        """
        Show a tooltip for the given url based on cached information.
        """
        cache = ImageTooltip.cache

        if url in cache:
            if not cache[url][0]:  # error, image was not found
                ImageTooltip.tooltipError(cache[url][1], pos)
            else:
                ImageTooltip.tooltip(cache[url][1], pos)
            return True
        return False

    def tooltipError(message, pos):
        """
        Display a tooltip with an error message at the given position.
        """
        ImageTooltip.editor.doTooltip(pos, message)

    def tooltip(image, pos):
        """
        Display a tooltip with an image at the given position.
        """
        px = image
        buffer = QBuffer()
        buffer.open(QIODevice.WriteOnly)
        px.save(buffer, "PNG", quality=100)
        image = bytes(buffer.data().toBase64()).decode()
        tt = "<p><img src='data:image/png;base64,{}'></p>".format(image)
        ImageTooltip.editor.doTooltip(pos, tt)



#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import QModelIndex
from PyQt5.QtCore import QSignalMapper
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QBrush
from PyQt5.QtGui import QStandardItem
from PyQt5.QtGui import QStandardItemModel
from PyQt5.QtWidgets import QAction, QMenu

from manuskript.enums import Plot
from manuskript.enums import PlotStep
from manuskript.functions import toInt, mainWindow


class plotModel(QStandardItemModel):
    def __init__(self, parent):
        QStandardItemModel.__init__(self, 0, 3, parent)
        self.setHorizontalHeaderLabels([i.name for i in Plot])
        self.mw = mainWindow()

        self.updatePlotPersoButton()

    ###############################################################################
    # QUERIES
    ###############################################################################

    def getPlotsByImportance(self):
        plots = [[], [], []]
        for i in range(self.rowCount()):
            importance = self.item(i, Plot.importance).text()
            ID = self.item(i, Plot.ID).text()
            plots[2 - toInt(importance)].append(ID)
        return plots

    def getSubPlotsByID(self, ID):
        index = self.getIndexFromID(ID)
        if not index.isValid():
            return
        index = index.sibling(index.row(), Plot.steps)
        item = self.itemFromIndex(index)
        lst = []
        for i in range(item.rowCount()):
            if item.child(i, PlotStep.ID):
                _ID = item.child(i, PlotStep.ID).text()

                # Don't know why sometimes name is None (while drag'n'dropping
                # several items)
                if item.child(i, PlotStep.name):
                    name = item.child(i, PlotStep.name).text()
                else:
                    name = ""

                # Don't know why sometimes summary is None
                if item.child(i, PlotStep.summary):
                    summary = item.child(i, PlotStep.summary).text()
                else:
                    summary = ""

                lst.append((_ID, name, summary))
        return lst

    def getPlotNameByID(self, ID):
        for i in range(self.rowCount()):
            _ID = self.item(i, Plot.ID).text()
            if _ID == ID or toInt(_ID) == ID:
                name = self.item(i, Plot.name).text()
                return name
        return None

    def getPlotImportanceByRow(self, row):
        for i in range(self.rowCount()):
            if i == row:
                importance = self.item(i, Plot.importance).text()
                return importance
        return "0" # Default to "Minor"

    def getSubPlotTextsByID(self, plotID, subplotRaw):
        """Returns a tuple (name, summary) for the subplot whose raw in the model
        is ``subplotRaw``, of plot whose ID is ``plotID``.
        """
        plotIndex = self.getIndexFromID(plotID)
        name = plotIndex.child(subplotRaw, PlotStep.name).data()
        summary = plotIndex.child(subplotRaw, PlotStep.summary).data()
        return name, summary

    def getIndexFromID(self, ID):
        for i in range(self.rowCount()):
            _ID = self.item(i, Plot.ID).text()
            if _ID == ID or toInt(_ID) == ID:
                return self.index(i, 0)
        return QModelIndex()

    def currentIndex(self):
        i = self.mw.lstPlots.currentIndex()
        if i.isValid():
            return i
        else:
            return None

    ###############################################################################
    # ADDING / REMOVING
    ###############################################################################

    def addPlot(self):
        p = QStandardItem(self.tr("New plot"))
        _id = QStandardItem(self.getUniqueID())
        importance = QStandardItem(str(0))
        self.appendRow([p, _id, importance, QStandardItem("Characters"),
                        QStandardItem(), QStandardItem(), QStandardItem("Resolution steps")])

    def getUniqueID(self, parent=QModelIndex()):
        """Returns an unused ID"""
        parentItem = self.itemFromIndex(parent)
        vals = []
        for i in range(self.rowCount(parent)):
            index = self.index(i, Plot.ID, parent)
            # item = self.item(i, Plot.ID)
            if index.isValid() and index.data():
                vals.append(int(index.data()))

        k = 0
        while k in vals:
            k += 1
        return str(k)

    def removePlot(self, index):
        self.takeRow(index.row())

    ###############################################################################
    # SUBPLOTS
    ###############################################################################

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                if section == PlotStep.name:
                    return self.tr("Name")
                elif section == PlotStep.meta:
                    return self.tr("Meta")
                else:
                    return ""
            else:
                return ""
        else:
            return QStandardItemModel.headerData(self, section, orientation, role)

    def data(self, index, role=Qt.DisplayRole):
        if index.parent().isValid() and \
                        index.parent().column() == Plot.steps and \
                        index.column() == PlotStep.meta:
            if role == Qt.TextAlignmentRole:
                return Qt.AlignRight | Qt.AlignVCenter
            elif role == Qt.ForegroundRole:
                return QBrush(Qt.gray)
            else:
                return QStandardItemModel.data(self, index, role)

        else:
            return QStandardItemModel.data(self, index, role)

    def addSubPlot(self):
        index = self.mw.lstPlots.currentPlotIndex()
        if not index.isValid():
            return

        parent = index.sibling(index.row(), Plot.steps)
        parentItem = self.item(index.row(), Plot.steps)

        if not parentItem:
            return

        p = QStandardItem(self.tr("New step"))
        _id = QStandardItem(self.getUniqueID(parent))
        summary = QStandardItem()

        currentIndex = self.mw.lstSubPlots.selectionModel().selectedIndexes()
        if currentIndex:
            # We use last item of selection in case of many
            currentIndex = currentIndex[-1]
            row = currentIndex.row() + 1
            parentItem.insertRow(row, [p, _id, QStandardItem(), summary])
            # Select last index
            self.mw.lstSubPlots.setCurrentIndex(currentIndex.sibling(row, 0))
        else:
            # Don't know why, if summary is in third position, then drag/drop deletes it...
            parentItem.appendRow([p, _id, QStandardItem(), summary])
            # Select last index
            self.mw.lstSubPlots.setCurrentIndex(parent.child(self.rowCount(parent) - 1, 0))

    def removeSubPlot(self):
        """
        Remove all selected subplots / plot steps, in mw.lstSubPlots.
        """
        parent = self.mw.lstSubPlots.rootIndex()
        if not parent.isValid():
            return
        parentItem = self.itemFromIndex(parent)

        while self.mw.lstSubPlots.selectionModel().selectedRows():
            i = self.mw.lstSubPlots.selectionModel().selectedRows()[0]
            parentItem.takeRow(i.row())

    def flags(self, index):
        parent = index.parent()
        if parent.isValid():  # this is a subitem
            return Qt.ItemIsEnabled | Qt.ItemIsEditable | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled
        else:
            return QStandardItemModel.flags(self, index)

    ###############################################################################
    # PLOT PERSOS
    ###############################################################################

    def addPlotPerso(self, v):
        index = self.mw.lstPlots.currentPlotIndex()
        if index.isValid():
            if not self.item(index.row(), Plot.characters):
                self.setItem(index.row(), Plot.characters, QStandardItem())

            item = self.item(index.row(), Plot.characters)

            # We check that the PersoID is not in the list yet
            for i in range(item.rowCount()):
                if item.child(i).text() == str(v):
                    return

            item.appendRow(QStandardItem(str(v)))

    def removePlotPerso(self):
        index = self.mw.lstPlotPerso.currentIndex()
        if not index.isValid():
            return
        parent = index.parent()
        parentItem = self.itemFromIndex(parent)
        parentItem.takeRow(index.row())

    def updatePlotPersoButton(self):
        menu = QMenu(self.mw)

        menus = []
        for i in [self.tr("Main"), self.tr("Secondary"), self.tr("Minor")]:
            m = QMenu(i, menu)
            menus.append(m)
            menu.addMenu(m)

        mpr = QSignalMapper(menu)
        for i in range(self.mw.mdlCharacter.rowCount()):
            a = QAction(self.mw.mdlCharacter.name(i), menu)
            a.setIcon(self.mw.mdlCharacter.icon(i))
            a.triggered.connect(mpr.map)
            mpr.setMapping(a, int(self.mw.mdlCharacter.ID(i)))

            imp = toInt(self.mw.mdlCharacter.importance(i))

            menus[2 - imp].addAction(a)

        # Disabling empty menus
        for m in menus:
            if not m.actions():
                m.setEnabled(False)

        mpr.mapped.connect(self.addPlotPerso)
        self.mw.btnAddPlotPerso.setMenu(menu)



#!/usr/bin/env python
# --!-- coding: utf8 --!--

"""Tests for settingsWindow"""

def test_loadExportWiget(MWSampleProject):
    """
    Simply tests that export widget loads properly.
    """
    MW = MWSampleProject

    # Loading from mainWindow
    MW.doCompile()
    E = MW.dialog
    assert E.isVisible()
    E.hide()

    # Load exporter manager
    E.openManager()
    EM = E.dialog
    assert EM.isVisible()
    EM.hide()

    EM.close()
    E.close()

#FIXME: test significant stuff



#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtWidgets import QWidget, QAbstractItemView

from manuskript.enums import Outline
from manuskript.ui.views.basicItemView_ui import Ui_basicItemView


class basicItemView(QWidget, Ui_basicItemView):
    def __init__(self, parent=None):
        QWidget.__init__(self)
        self.setupUi(self)
        self.txtSummarySentence.setColumn(Outline.summarySentence)
        self.txtSummaryFull.setColumn(Outline.summaryFull)
        self.txtGoal.setColumn(Outline.setGoal)

    def setModels(self, mdlOutline, mdlCharacter, mdlLabels, mdlStatus):
        self.cmbPOV.setModels(mdlCharacter, mdlOutline)
        self.txtSummarySentence.setModel(mdlOutline)
        self.txtSummaryFull.setModel(mdlOutline)
        self.txtGoal.setModel(mdlOutline)

    def getIndexes(self, sourceView):
        """Returns a list of indexes from list of QItemSelectionRange"""
        indexes = []

        for i in sourceView.selection().indexes():
            if i.column() != 0:
                continue

            if i not in indexes:
                indexes.append(i)

        return indexes

    def selectionChanged(self):
        if isinstance(self.sender(), QAbstractItemView):
            selectionModel = self.sender().selectionModel()
        else:
            selectionModel = self.sender()

        indexes = self.getIndexes(selectionModel)

        if len(indexes) == 0:
            self.setEnabled(False)

        elif len(indexes) == 1:
            self.setEnabled(True)
            idx = indexes[0]
            self.txtSummarySentence.setCurrentModelIndex(idx)
            self.txtSummaryFull.setCurrentModelIndex(idx)
            self.cmbPOV.setCurrentModelIndex(idx)
            self.txtGoal.setCurrentModelIndex(idx)

        else:
            self.setEnabled(True)
            self.txtSummarySentence.setCurrentModelIndexes(indexes)
            self.txtSummaryFull.setCurrentModelIndexes(indexes)
            self.cmbPOV.setCurrentModelIndexes(indexes)
            self.txtGoal.setCurrentModelIndexes(indexes)

    def setDict(self, d):
        self.txtSummaryFull.setDict(d)

    def toggleSpellcheck(self, v):
        self.txtSummaryFull.toggleSpellcheck(v)
